from typing import Optional, List
from msgspec import Struct, field

from .messagesData import MessageData
from .deltas.custom_type import Value




class addedParticipant(Struct, frozen=True, eq=False):
    """Adeed participant's information."""
    name: str = field(name="fullName")
    """added participant's name"""
    user_id: str = field(name="userFbId")
    """added participant's user id"""
    last_unsubscribed: Optional[str] = field(name="lastUnsubscribeTimestampMs", default=None)
    """join timestamp"""


################ All main classes ##################

class ApprovalMode(Struct, frozen=True, eq=False, tag="ApprovalMode", tag_field="class"):
    """Thread's Updated Approval mode data."""
    mode: str
    """mode wether Approval Mode is on 'APPROVALS'(on) or OPEN' (off)"""
    messageMetadata: MessageData 
    """Extra information about the message."""


class ApprovalQueue(Struct, frozen=True, eq=False, tag="ApprovalQueue", tag_field="class"):
    """A user's join request or disapproved join request's data exclude."""
    requester_id: str = field(name="recipientFbId")
    """Requested `User`'s user id"""
    action: str 
    """action can be either 'REQUESTED' or 'REMOVED'."""
    messageMetadata: MessageData
    """Extra message information inluding message author, timestamp, stamp"""
    inviter_id: Optional[str] = field(name="inviterFbId", default=None)
    """The participant who added another User to the group."""
    request_timestamp: Optional[str] = field(name="requestTimestamp", default=None)
    """Request timestamp"""
    requestSource: str = "JOIN_THROUGH_LINK"
    """Wether joined through Link (default value) or Added by another group participant if added then value will be 'ADD'"""


class JoinableMode(Struct, frozen=True, eq=False, tag="JoinableMode", tag_field="class"):
    """Joining through Link to a Group."""
    mode: str   # JOINABLE / PRIVATE
    """Can be joined through Link if value 'JOINABLE' and can't if value is 'PRIAVTE'"""
    messageMetadata: MessageData
    """Extra message information."""


class ParticipantsAdded(Struct, frozen=True, eq=False, tag="ParticipantsAddedToGroupThread", tag_field="class"):
    """Added Participant's information"""
    added_participants: List[addedParticipant] = field(name="addedParticipants")
    """List of 'addedParticipant' object contains added participants data."""
    messageMetadata: MessageData
    """Extra information such as author (The one who added the Users), timestamp etc."""
    participants: tuple = tuple()
    """Tuple of all participants Id in the Group."""

class ParticipantLeft(Struct, frozen=True, eq=False, tag="ParticipantLeftGroupThread", tag_field="class"):
    """Left participant's information"""
    left_participant: str = field(name="leftParticipantFbId")
    """The Id of the left participant."""
    messageMetadata: MessageData
    """Extra message information."""
    participants: tuple = tuple()
    """Tuple of all participants Id in the Group."""

class AdminRemoved(Struct, frozen=True, eq=False, tag="AdminRemovedFromGroupThread", tag_field="class"):
    """Removed Admin event related data."""
    removed_admins: List[str] = field(name="removedAdminFbIds")
    """List of removed admins ids"""
    messageMetadata: MessageData
    """The message metadata information"""

class ThreadName(Struct, frozen=True, eq=False, tag="ThreadName", tag_field="class"):
    """Updated Thread's name"""
    name: str
    """New name of the Thread"""
    messageMetadata: MessageData
    """Metadata of the message"""
    participants: tuple = tuple()
    """Tuple of all participants Id in the Group."""


# Payload class 
class ApprovedUser(Struct, frozen=True, eq=False):
    """Approved User event data"""
    thread_id: Value = field(name="threadKey")
    """The thread Id where the User was accepted."""
    contact_id: int = field(name="contactId")
    """The Admin Id who approved The User to Thread"""
    approved_user_id: int = field(name="subscribe_actor_id")
    """The approved User Id"""

# payload class
class ChangeViwerStatus(Struct, frozen=True, eq=False):
    """Changed Viewer Status information"""
    user_id: str = field(name="actorFbid")
    """The User (Client) who changed the View"""
    thread_id: Value = field(name="threadKey")
    """The Thread's (User Thread) Id which view was changed"""
    can_reply: bool = field(name="canViewerReply")
    """User can reply to that Thread if `True` else it is `False`"""
    reason: int  # 0 for fb and 2 for messenger block
    """0 if the User blocked the Thread (User Thread in that case) in Facebook and 2 if in Messenger"""
    is_messenger_blocked: Optional[bool] = field(name="isMsgBlockedByViewer", default=None)
    """If User is blocked on Messenger"""
    messenger_blocked_timestamp: Optional[int] = field(name="isMsgBlockedTimestamp", default=None)
    """The block event timestamp"""
    is_facebook_blocked: Optional[bool] = field(name="isFBBlockedByViewer", default=None)
    """If the User is blocked in Facebook."""
    facebook_blocked_timestamp: Optional[int] = field(name="isFBBlockedTimestamp", default=None)
    """Facebook block timestamp"""


class MuteThread(Struct, frozen=True, eq=False):
    """Thread Mute information"""
    thread_id: Value = field(name="threadKey")
    """Muted Thread's Id"""
    mute_until: int = field(name="muteUntil")
    """Mute duration"""


class ThreadMuteSettings(Struct, frozen=True, eq=False, tag="ThreadMuteSettings", tag_field="class"):
    """Thread mute related information."""
    user_id: str = field(name="actorFbId")
    """The User (Client because other's mute information is not recevied) who muted the Thread """
    thread_id: Value = field(name="threadKey")
    """The Threas that was muted"""
    expire_time: int = field(name="expireTime")
    """Mute expire time."""

class ThreadAction(Struct, frozen=True, eq=False, tag="ThreadAction", tag_field="class"):
    """information of the Archived from Inbox."""
    action: str
    """The action that was done"""
    thread_id: Value = field(name="threadKey")
    """The Thread action was done on."""


class ThreadFolderMove(Struct, frozen=True, eq=False, tag="ThreadFolder", tag_field="class"):
    """Moved Thread folder data."""
    user_id: str 
    """The user who moved the Thread"""
    folder: str 
    """To the folder Location it was moved such as Inbox, Archive, etc."""
    thread_id: Value = field(name="threadKey")
    """The Thread that was moved."""

class ThreadDelete(Struct, frozen=True, eq=False, tag="ThreadDelete", tag_field="class"):
    """Deleted Thread information."""
    user_id: str = field(name="actorFbId") 
    """The User who deleted the Thread."""
    thread_ids: List[Value] = field(name="threadKeys")
    """List of Threas Ids that were deleted."""






# ----------------- Only  Group Admins action Related Delta -------- #

class AdminAdded(Struct, frozen=True, eq=False):
    """
    Added admin's information. 
    """

    aded_admin: str = field(name="TARGET_ID")
    """Removed/Added admin Id"""
    thread_type: str = field(name="THREAD_CATEGORY")
    """Type of the thread"""
    # admin_event: str = field(name="ADMIN_EVENT")


class ThreadMessagePin(Struct, frozen=True, eq=False):
    """Pinned Message"""
    message_id: str = field(name="pinned_message_id")
    """The pinned message's id"""

class ThreadMessageUnPin(Struct, frozen=True, eq=False):
    """Unpinned message"""
    message_id: str =  field(name="pinned_message_id")
    """The Unpinned message's id"""

class ThreadMessageSharing(Struct, frozen=True, eq=False):
    """Threas Message Sharing Mode information"""
    mode: str = field(name="limit_sharing_type")
    """Message mode is on if value is 'enabled' and off if value is 'disabled'"""
    sender_name: str 
    """The User's name who changed Message Sharing mode"""
    sender_id: str 
    """The User's Id"""

class ThreadMagicWord(Struct, frozen=True, eq=False):
    """Updated Magic word information."""
    magic_word: str 
    """The name of the Magic Word"""
    theme_name: str 
    """theme_name of the Magic word"""
    emoji: str = field(name="emoji_effect")
    """The emoji used for the Magic word."""
    removed_magic_word_count: str 
    """0 if not removed any Magic Word otherwise greater than 0"""
    new_magic_word_count: str
    """0 if not added any Magic Word otherwise greater than 0"""

class ThreadNickname(Struct, frozen=True, eq=False):
    """Changed nickname information of a participant"""
    nickname: str 
    """The new nickname of the User"""
    participant_id: str
    """The changed nickname User Id."""

class ThreadTheme(Struct, frozen=True, eq=False):
    """Updated Threaf theme information"""
    theme_id: str 
    """Id of the theme"""
    theme_name: str = field(name="theme_name_with_subtitle")
    """Name of the Theme"""
    theme_emoji: str 
    """The Quick Reaction Emoji for the Theme"""
    theme_type: str
    """Type of the theme"""
    theme_color: str
    """The color of the theme."""
    gradient: str
    """The gradient color of the theme"""
    accessibility_label: str
    """The keyword usee to label the theme."""

class ThreadEmoji(Struct, frozen=True, eq=False):
    """Updated Threaf Emoji information"""
    emoji: str = field(name="thread_quick_reaction_emoji")
    """The new quick reaction emoji"""
    emoji_url: str = field(name="thread_quick_reaction_emoji_url")
    """The url of the emoji"""


#--------------- End ------------------- #



class ForcedFetch(Struct, frozen=True, eq=False, tag="ForcedFetch", tag_field="class"):
    """Usually recevied when messenger group cover photo changes or along with other deltas"""
    threadKey: Value
    messageId: Optional[str] = None
    """messageId only recevied Usually when a Thread (Group) Picture is changed"""
    type: str = "None"



""" Facebook Messenger uses Mqtt protocol to send and receive messages. This File handles Mqtt broker. """

from __future__ import annotations


import aiomqtt, asyncio, aiohttp
import random
import json
import time
import re

from asyncio import Task
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Optional
from yarl import URL


from .exception.errors import *

from .graphql import from_doc_id
from .state import State
from .logging.logger import get_logger
from .utils.utils import generate_uuid

logger = get_logger()

def generate_session_id()-> int:
    return random.randint(1, 2 ** 53)


def get_cookie_header(session: aiohttp.ClientSession, url: str) -> str:
    return session.cookie_jar.filter_cookies(URL(url)).output(header="", sep=";").lstrip()


def get_random_reconnect_time() -> int:
    """Generate random reconnection time between 26-60 minutes"""
    min_time = 26 * 60 * 1000  # 26 minutes in milliseconds
    max_time = 60 * 60 * 1000  # 60 minutes in milliseconds
    return random.randint(min_time, max_time)


@dataclass(slots=True)
class Mqtt:
    _state: State = field()
    _mqttClient: aiomqtt.Client = field()
    _chat_on: bool = field()
    _foreground: bool = field()
    _sequence_id: int = field()

    _mqttClientID: str = field()
    _mqttAppID: str = field()
    _region: str = field()
    _sync_token: Any = field(default=None)

    _update_presence: bool = field(default=True)
    _auto_reconnect: bool = field(default=True)
    _presence_task: Optional[Task] = field(default=None)
    _reconnect_task: Optional[Task] = field(default=None)
    _listen_task: Optional[Task] = field(default=None)
    _running: bool = field(default=False)
    _reconnecting: bool = field(default=False)

    _message_handler: Optional[Callable] = field(default=None)

    HOST = "edge-chat.facebook.com" # Mqtt host for facebook
    iris_re   = re.compile(rb'"irisSeqId":\s*"(\d+)"')
    first_re  = re.compile(rb'"firstDeltaSeqId":\s*(\d+)')
    last_re   = re.compile(rb'"lastIssuedSeqId":\s*(\d+)')
    tq_re     = re.compile(rb'"tqSeqId":\s*"(\d+)"')
    sync_re   = re.compile(rb'"syncToken":\s*"([^"]+)"')

    
    @classmethod
    async def connect(
            cls, 
            state: State, 
            chat_on: bool, 
            foreground: bool,
            message_handler: Callable,
            update_presence: bool = True,
            auto_reconnect: bool = True
            )-> Mqtt:

        # configuring mqtt client 
        mqttClint = aiomqtt.Client(
            hostname=cls.HOST,
            identifier="mqttwsclient",
            clean_session=True,
            protocol=aiomqtt.ProtocolVersion.V31,
            transport="websockets",
            port=443,
            keepalive=60
        )

        sequence_id = await cls._fetch_sequence_id(state)
        
        # creating `Mqtt` class instance for handlling mqtt connections
        self = cls(
            _state=state,
            _mqttClient=mqttClint,
            _chat_on=chat_on,
            _foreground=foreground,
            _sequence_id=sequence_id,
            _mqttClientID=state._mqttClientID,
            _mqttAppID=state._mqttAppID,
            _region=state._region,  
            _message_handler=message_handler,
            _update_presence=update_presence,
            _auto_reconnect=auto_reconnect
        )
    
        self._mqttClient._client.tls_set()
        self._configure_mqtt_options()

        # connect the mqtt client
        await self._mqttClient.__aenter__()

        if self._mqttClient._client.is_connected():
            logger.info("🔨 Successfully connected to MQTT broker")
            await self._messenger_queue_publish()
            self._running = True

            # run them in the background
            if self._update_presence:
                self._presence_task = asyncio.create_task(self._presence_updater())
            if self._auto_reconnect:
                self._reconnect_task = asyncio.create_task(self._schedule_reconnect())

            self._listen_task = asyncio.create_task(self.listen())

        return self


    @staticmethod
    async def _fetch_sequence_id(state)-> int:
        """Fetch sequence ID."""
        params = {
            "limit": 1,
            "tags": ["INBOX"],
            "before": None,
            "includeDeliveryReceipts": False,
            "includeSeqID": True,
            }
        j = await state._graphql_requests(from_doc_id("1349387578499440", params))
        sequence_id = j[0]["viewer"]["message_threads"]["sync_sequence_id"] #type: ignore
        logger.debug(f"fetched sequence id: {sequence_id}")
        return int(sequence_id)


    def _configure_mqtt_options(self)-> None:
        session_id = generate_session_id()
        region = self._region
        mqttClientID = self._mqttClientID
        mqttAppID = self._mqttAppID

        # The topics fbchat-muqit will listen to 
        # with open("./fbchat_muqit/topics.json", "r") as f:
        #     data = json.loads(f.read())
        # topics = [i for i in data]
        # print(*topics)
        topics = [
            "/legacy_web",  # web related messages during graphql requests
            "/ls_req",    # used to publish message payloads
            "/ls_resp",   # receives response after pub to /ls_req
            "/t_ms",      # all kinds of message, message events received
            "/rtc_multi",
            #"/t_rtc_multi",  # receives group calls
            "/thread_typing", # typing status update received
            "/orca_typing_notifications",  # Messenger notifi.
            "/orca_presence",     # receives users presence updates
            "/br_sr",
            "/friend_request",      # receives friend request notification
            "/friending_state_change", # when friend request confirmed/removed
            "/friend_requests_seen",  # new friend request seen
            "/sr_res",
            "/webrtc",
            "/onevc",
            "/notify_disconnect",
            "/mercury",
            "/inbox",
            "/messaging_events",
            "/orca_message_notifications",
            "/pp",
            "/webrtc_response",
        ]
        username = {
            "u": self._state.user_id,
            "s": session_id,
            "chat_on": self._chat_on,
            "fg": self._foreground,
            "d": mqttClientID,
            "aid": mqttAppID,
            "st": topics,
            "pm": [],
            "cp": 3,
            "ecp": 10,
            "ct": "websocket",
            "mqtt_sid": "",
            "dc": "",
            "no_auto_fg": True,
            "gas": None,
            "pack": [],
            "p": None,
            "aids": None,
            "a": self._state._userAgent
        }
        self._mqttClient._client.username_pw_set(json.dumps(username))

        headers = {
            "Cookie": get_cookie_header(
                self._state._session, "https://edge-chat.facebook.com/chat"
            ),
            "User-Agent": self._state._userAgent,
            "Origin": "https://www.facebook.com",
            "Host": self.HOST
        }

        self._mqttClient._client.ws_set_options(
            path=f"/chat?region={region}&sid={session_id}&cid={mqttClientID}", headers=headers
        )

    
    async def _messenger_queue_publish(self)-> None:
        # configure receiving messages.
        payload = {
            "sync_api_version": 10,
            "max_deltas_able_to_process": 1000,
            "delta_batch_size": 500,
            "encoding": "JSON",
            "entity_fbid": self._state.user_id,
        }
        if self._sync_token is None:
            topic = "/messenger_sync_create_queue"
            payload["initial_titan_sequence_id"] = str(self._sequence_id)
            payload["device_params"] = None
        else:
            topic = "/messenger_sync_get_diffs"
            payload["last_seq_id"] = str(self._sequence_id)
            payload["sync_token"] = self._sync_token

        await self._mqttClient.publish(topic, json.dumps(payload), qos=1)

    def extract_meta(self, raw):
        """extracts sequence and sync token"""
        ids = {
            "firstDeltaSeqId": int(m.group(1)) if (m := self.first_re.search(raw)) else None,
            "lastIssuedSeqId": int(m.group(1)) if (m := self.last_re.search(raw)) else None,
            "syncToken": m.group(1).decode() if (m := self.sync_re.search(raw)) else None,
        }
        self._sequence_id = ids["lastIssuedSeqId"] or ids["firstDeltaSeqId"] or self._sequence_id 
        self._sync_token = ids["syncToken"] or self._sync_token
        if not any(ids.values()):
            logger.debug("No seqId/syncToken found in payload slice")

    

    async def listen(self):
        if not self._mqttClient:
            raise FBChatError("Mqtt Client is not initialised cannot start listening!")
        logger.info("Starting MQTT listening loop to listen to events")
        try:
            async for message in self._mqttClient.messages:
                if not self._running and not self._reconnecting:
                    break
                try:    
                    if b'lastIssuedSeqId' in message.payload or b'firstDeltaSeqId' in message.payload or b'syncToken' in message.payload: #type: ignore
                        self.extract_meta(message.payload)


                    payload = message.payload #type: ignore
                    topic = message.topic.value
                    if self._message_handler:
                        await self._message_handler(topic, payload)
                except Exception as e:
                    FBChatError("Failed to receive message payloads", original_exception=e)

        except asyncio.CancelledError:
            # normal shutdown, just exit silently
            logger.debug("MQTT listening loop cancelled")
            return
        except Exception as e:
            logger.error(f"MQTT listening loop error: {e}")
            if self._running and self._auto_reconnect:
                logger.info("Attempting to reconnect...")
                await self._reconnect()
        finally:
            logger.info("MQTT listening loop ended")


    def parse_json(self, data):
        try:
            return json.loads(data)
        except json.JSONDecodeError as e:
            logger.error(f"Failed to parse message payload as JSON: {e}")
            return {"raw_data": data}  # Return raw data instead of raising
        except Exception as e:
            logger.error(f"Unexpected error parsing JSON: {e}")
            return {"error": str(e), "raw_data": data}


    async def set_foreground(self, value):
        payload = json.dumps({"foreground": value})
        await self._mqttClient.publish("/foreground_state", payload=payload, qos=1)
        self._foreground = value


    async def set_chat_on(self, value):
        data = {"make_user_available_when_in_foreground": value}
        payload = json.dumps(data)
        await self._mqttClient.publish("/set_client_settings", payload=payload, qos=1)
        self._chat_on = value


    def _generate_presence(self) -> Dict[str, Any]:
        """Generate presence payload"""
        return {
            "user_id": self._state.user_id,
            "last_active": int(time.time() * 1000),
            "active": True
        }


    async def _presence_updater(self):
        """Periodically update presence status"""
        while self._running:
            try:
                if self._mqttClient._client.is_connected():
                    presence_payload = self._generate_presence()
                    await self._mqttClient.publish(
                        '/orca_presence', 
                        json.dumps({"p": presence_payload}), 
                        qos=1
                    )
                    logger.debug("Presence updated")
                await asyncio.sleep(50)  # Update every 50 seconds

            except asyncio.CancelledError:
                logger.debug("Presence updater cancelled")
                return
            except Exception as e:
                logger.error(f"Error updating presence: {e}")
                await asyncio.sleep(10)


    
    async def _cancel_task(self, task: asyncio.Task | None):
        """Cancel an asyncio task if running, and await its cleanup."""
        if task and not task.done():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass


    async def stop(self):
        """Clean disconnect from MQTT"""
        self._running = False  # Stop background tasks
        await self._cancel_task(self._listen_task)
        await self._cancel_task(self._reconnect_task)
        await self._cancel_task(self._presence_task)
        # Disconnect MQTT client
        if self._mqttClient:
            try:
                await self._mqttClient.__aexit__(None, None, None)
                logger.info("Disconnected from MQTT")
            except Exception as e:
                logger.error(f"Error disconnecting: {e}")

    ###### mqtt reconnection Scheduled ######
    async def _reconnect(self):
        """Perform reconnection"""
        self._reconnecting = True

        await self._cancel_task(self._listen_task)
        await self._cancel_task(self._presence_task)
        # Disconnect current client
        if self._mqttClient:
            await self._mqttClient.__aexit__(None, None, None)
       
        await asyncio.sleep(2)
        
        self._mqttClientID = self._state._mqttClientID = generate_uuid()
        logger.debug(f"Generated new MQTT client ID: {self._mqttClientID}")
        self._sequence_id = await self._fetch_sequence_id(self._state)
        logger.debug(f"Fetched new sequence id: {self._sequence_id}")

        self._mqttClient = aiomqtt.Client(
            hostname=self.HOST,
            identifier="mqttwsclient",
            clean_session=True,
            protocol=aiomqtt.ProtocolVersion.V31,
            transport="websockets",
            port=443,
            keepalive=60
        )
        
        # Reconnect with new configuration
        self._mqttClient._client.tls_set()
        self._configure_mqtt_options()
        await self._mqttClient.__aenter__()
        
        if self._mqttClient._client.is_connected():
            logger.info("✅ MQTT reconnected successfully")
            self._sync_token = None

            logger.debug("Publishing messengee queue...")
            await self._messenger_queue_publish()
            logger.debug("Publishing messengee queue...")

            self._listen_task = asyncio.create_task(self.listen())
            self._presence_task = asyncio.create_task(self._presence_updater())
            logger.info("✅ Reconnected and restarted listen/presence loops")
        self._reconnecting = False


    async def _schedule_reconnect(self):
        """Schedule periodic reconnections with random intervals"""
        while self._running:
            reconnect_time = get_random_reconnect_time() / 1000  # Convert to seconds
            logger.info(f"Scheduled reconnect in {int(reconnect_time / 60)} minutes...")
            await asyncio.sleep(reconnect_time)
            logger.info("Reconnecting MQTT with new clientID...")
            try:
                await self._reconnect()
            except Exception as e:
                logger.error(f"Reconnection failed: {e}")






# Facebook Realtime WebSocket connection for notifications, friend requests, and presence updates

from __future__ import annotations

import asyncio
import struct
import aiohttp
from dataclasses import dataclass, field
from typing import Optional, Dict, Any, List, Callable
from urllib.parse import urlencode
from yarl import URL

from .state import State
from .logging.logger import get_logger
from .exception.errors import FBChatError
@dataclass 
class NotificationEvent:
    type: str 
    notif_id: str
    body: str | None 
    sender_id: str
    url: str | None
    timestamp: int
    seen_state: Any 
    raw_data: Any


logger = get_logger()


class RealtimeEventEmitter:
    """Event emitter for realtime WebSocket events"""
    
    def __init__(self):
        self._listeners: Dict[str, List[Callable]] = {}
        self._running = True
    
    def on(self, event: str, callback: Callable):
        """Register an event listener"""
        if event not in self._listeners:
            self._listeners[event] = []
        self._listeners[event].append(callback)
        logger.debug(f"Registered listener for event: {event}")
    
    def off(self, event: str, callback: Callable):
        """Remove an event listener"""
        if event in self._listeners:
            try:
                self._listeners[event].remove(callback)
                if not self._listeners[event]:
                    del self._listeners[event]
            except ValueError:
                pass
    
    async def emit(self, event: str, *args, **kwargs):
        """Emit an event to all listeners"""
        if not self._running or event not in self._listeners:
            return
            
        for callback in self._listeners[event][:]:
            try:
                if asyncio.iscoroutinefunction(callback):
                    await callback(*args, **kwargs)
                else:
                    callback(*args, **kwargs)
            except Exception as e:
                logger.error(f"Error in event listener for {event}: {e}")
                if event != 'error':
                    await self.emit('error', e)
    
    def stop(self):
        """Stop the event emitter"""
        self._running = False
        self._listeners.clear()

def format_notification(data: Dict[str, Any]) -> Optional[NotificationEvent]:
    """Format notification data from WebSocket response"""
    try:
        if not data.get('data') or not data['data'].get('viewer'):
            return None
        
        notif_edge = (data['data']['viewer']
                     .get('notifications_page', {})
                     .get('edges', [{}])[1:2])
        
        if not notif_edge:
            return None
            
        notif = notif_edge[0].get('node', {}).get('notif')
        if not notif:
            return None
        
        # Extract sender ID from tracking
        tracking = notif.get('tracking', {})
        from_uids = tracking.get('from_uids', {})
        sender_id = next(iter(from_uids.keys())) if from_uids else None
        
        return NotificationEvent(
            type="notification",
            notif_id=notif.get('notif_id'),
            body=notif.get('body', {}).get('text') if notif.get('body') else None,
            sender_id=sender_id or "",
            url=notif.get('url'),
            timestamp=notif.get('creation_time', {}).get('timestamp'),
            seen_state=notif.get('seen_state'),
            raw_data=data
        )
        
    except Exception as e:
        logger.error(f"Error formatting notification: {e}")
        return None

def get_cookie_header(session: aiohttp.ClientSession, url: str) -> str:
    """Extract cookies for the given URL"""
    return session.cookie_jar.filter_cookies(URL(url)).output(header="", sep=";").lstrip()

@dataclass
class FacebookRealtime:
    """Facebook WebSocket realtime connection handler"""
    
    _state: State = field()
    _ws: Optional[aiohttp.ClientWebSocketResponse] = field(default=None)
    _session: Optional[aiohttp.ClientSession] = field(default=None)
    _emitter: RealtimeEventEmitter = field(default_factory=RealtimeEventEmitter)
    _listen_task: Optional[asyncio.Task] = field(default=None)
    _running: bool = field(default=False)
    _message_handler: Optional[Callable] = field(default=None)
    
    # WebSocket connection parameters
    _WS_HOST = "wss://gateway.facebook.com/ws/realtime"
    _APP_ID = "2220391788200892"
    
    @classmethod
    async def connect(cls, state: State, message_handler: Callable) -> FacebookRealtime:
        """Create and connect to Facebook realtime WebSocket"""
        instance = cls(_state=state, _message_handler=message_handler)
        await instance._connect()
        return instance
    
    def _get_subscriptions(self) -> List[str]:
        """Get list of subscription payloads for different Facebook services"""
        user_id = self._state.user_id
        
        return [
            '{"x-dgw-app-XRSS-method":"Falco","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:USER_ACTIVITY_UPDATE_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"9525970914181809","x-dgw-app-XRSS-routing_hint":"UserActivitySubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:ACTOR_GATEWAY_EXPERIENCE_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"24191710730466150","x-dgw-app-XRSS-routing_hint":"CometActorGatewayExperienceSubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}',
            f'{{"x-dgw-app-XRSS-method":"FBLQ:comet_notifications_live_query_experimental","x-dgw-app-XRSS-doc_id":"9784489068321501","x-dgw-app-XRSS-actor_id":"{user_id}","x-dgw-app-XRSS-page_id":"{user_id}","x-dgw-app-XRSS-request_stream_retry":"false","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:FRIEND_REQUEST_CONFIRM_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"9687616244672204","x-dgw-app-XRSS-routing_hint":"FriendingCometFriendRequestConfirmSubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:FRIEND_REQUEST_RECEIVE_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"24047008371656912","x-dgw-app-XRSS-routing_hint":"FriendingCometFriendRequestReceiveSubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:RTWEB_CALL_BLOCKED_SETTING_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"24429620016626810","x-dgw-app-XRSS-routing_hint":"RTWebCallBlockedSettingSubscription_CallBlockSettingSubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}',
            '{"x-dgw-app-XRSS-method":"PresenceUnifiedJSON","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com/friends"}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:MESSENGER_CHAT_TABS_NOTIFICATION_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"23885219097739619","x-dgw-app-XRSS-routing_hint":"MWChatTabsNotificationSubscription_MessengerChatTabsNotificationSubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com/friends"}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:BATCH_NOTIFICATION_STATE_CHANGE_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"30300156509571373","x-dgw-app-XRSS-routing_hint":"CometBatchNotificationsStateChangeSubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com/friends"}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:NOTIFICATION_STATE_CHANGE_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"23864641996495578","x-dgw-app-XRSS-routing_hint":"CometNotificationsStateChangeSubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}',
            '{"x-dgw-app-XRSS-method":"FBGQLS:NOTIFICATION_STATE_CHANGE_SUBSCRIBE","x-dgw-app-XRSS-doc_id":"9754477301332178","x-dgw-app-XRSS-routing_hint":"CometFriendNotificationsStateChangeSubscription","x-dgw-app-xrs-body":"true","x-dgw-app-XRS-Accept-Ack":"RSAck","x-dgw-app-XRSS-http_referer":"https://www.facebook.com"}'
        ]
    
    async def _connect(self):
        """Establish WebSocket connection to Facebook realtime gateway"""
        try:
            # Build query parameters
            query_params = {
                "x-dgw-appid": self._APP_ID,
                "x-dgw-appversion": "0",
                "x-dgw-authtype": "1:0",
                "x-dgw-version": "5",
                "x-dgw-uuid": str(self._state.user_id),
                "x-dgw-tier": "prod",
                "x-dgw-deviceid": self._state._mqttClientID,
                "x-dgw-app-stream-group": "group1"
            }
            
            url = f"{self._WS_HOST}?{urlencode(query_params)}"
            
            # Get cookies
            cookies = get_cookie_header(self._state._session, "https://www.facebook.com")
            
            # Build headers
            headers = {
                "Cookie": cookies,
                "Origin": "https://www.facebook.com",
                "User-Agent": self._state._userAgent,
                "Referer": "https://www.facebook.com",
                "Host": URL(url).host,
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "en-US,en;q=0.9"
            }
            
            logger.debug(f"📤 Connecting to Facebook realtime WebSocket: {url}")
            # logger.debug(f"Headers: {headers}")
            
            # Create WebSocket session
            self._session = self._state._session
            self._ws = await self._session.ws_connect(url, headers=headers, heartbeat=20, autoping=True)
            
            logger.info("Connected to Facebook realtime WebSocket")
            
            # Send subscriptions
            await self._send_subscriptions()
            
            # Start background tasks
            self._running = True
            self._listen_task = asyncio.create_task(self._listen_loop())
            
            await self._emitter.emit('connected')
            
        except Exception as e:
            logger.error(f"💥 Connection error: {e}")
            await self._emitter.emit('error', e)
            # Schedule reconnection
            asyncio.create_task(self._schedule_reconnect())
    
    async def _send_subscriptions(self):
        """Send subscription messages to Facebook gateway"""
        subscriptions = self._get_subscriptions()
        
        logger.debug(f"📤 Sending subscription...")
        for index, payload in enumerate(subscriptions):
            try:
                # Format: [14, index, 0, payload_length] + payload + [0, 0]
                prefix = struct.pack('BBBi', 14, index, 0, len(payload))
                suffix = struct.pack('BB', 0, 0)
                full_message = prefix + payload.encode('utf-8') + suffix
                if self._ws: 
                    await self._ws.send_bytes(full_message)
                
            except Exception as e:
                logger.error(f"Error sending subscription {index}: {e}")
                if self._running:
                    logger.debug("Attempting to reconnect realtime..")
                    await self.reconnect()
        logger.debug(f"Successfully sent subscriptions!")
    
    async def _listen_loop(self):
        """Main message listening loop"""
        if not self._ws:
            raise FBChatError("Websocket not initilised for listening to realtime events")
        try:
            async for msg in self._ws:
                if self._message_handler:
                    await self._message_handler(msg.type, msg.data)
        except asyncio.CancelledError:
            logger.debug("🛑 Realtime loop cancelled")
            return
        except Exception as e:
            logger.error(f"Error in listen loop: {e}")
            await self._emitter.emit('error', e)

    async def listen(self):
        if self._listen_task:
            await self._listen_task

    def set_message_handler(self, handler: Callable):
        """Set the message handler function"""
        self._message_handler = handler


    async def _schedule_reconnect(self):
        """Schedule reconnection after a delay"""
        if not self._running:
            return
            
        logger.warning("🔌 Scheduling reconnection in 1 second...")
        await asyncio.sleep(1)
        
        if self._running:
            await self.reconnect()


    async def reconnect(self):
        await self._cleanup()
        await self._connect()

    
    async def _cleanup(self):
        """Clean up resources"""
        # Close WebSocket
        if self._ws and not self._ws.closed:
            await self._ws.close()
        
        # Close _session
        self._session = None


    def on(self, event: str, callback: Callable):
        """Register event listener"""
        self._emitter.on(event, callback)
    
    def off(self, event: str, callback: Callable):
        """Remove event listener"""
        self._emitter.off(event, callback)
    
    async def stop(self):
        """Stop the realtime connection"""
        logger.debug("Stopping Facebook realtime connection...")
        self._running = False
        self._emitter.stop()
        
        # Cancel reconnect task
        if self._listen_task and not self._listen_task.done():
            self._listen_task.cancel()
            try:
                await self._listen_task
            except asyncio.CancelledError:
                pass
        
        await self._cleanup()
        logger.info("Facebook realtime connection stopped")




from msgspec import Struct, field


class Typing(Struct, frozen=True, eq=False):
    """Typing information of the Typer"""
    sender_id: int = field(name="sender_fbid")
    """The Id of the User who is typing"""
    state: int
    """1 for typing status and 0 for not typing."""
    thread_id: str = field(name="thread", default="")
    """Thw thread Id where user is typing if it is empty then It is one-to-one so sender_id and thread_id is same."""



# pyright: reportUnknownVariableType=false
# pyright: reportUnknownMemberType=false
from msgspec import Struct, json
from typing import Any, Optional



class MessageSearchResult(Struct, frozen=True, eq=False):
    query: str
    """The Searched text"""
    result_index: int
    """Index of the query text"""
    thread_id: str
    """Thread Id of the message"""
    sender_name: str
    """The sender of the message"""
    message_id: str
    """Id of the message where query text was found."""
    timestamp_ms: int
    """timestamp of the message"""
    snippet: str
    """The message's text"""
    profile_pic_url: str 
    """message sender profile picture."""
    highlight_offset: str
    """offset of the query text in the message"""
    highlight_length: str
    """length of the highlighted query text in the message"""

class MessageSearchStatus(Struct, frozen=True, eq=False):
    query: str
    """query text"""
    total_results: int
    """total results found"""
    status: int
    """status of search result"""
    has_more: bool
    """If more messages has the query text or not"""
    next_cursor: Optional[str]
    """next message count probably not sure"""
    thread_id: str
    """Thread Id where query text was Searched"""

def parse_message_search(data: str):
    """Parse message search response from Facebook Messenger."""
    inner = json.decode(data)
    results = []
    status = None

    def decode_value(v)-> Any:
        """Decode Facebook's encoded values."""
        if isinstance(v, list) and len(v) >= 2 and isinstance(v[0], int):
            if v[0] == 19:  # String type
                return v[1]
            if v[0] == 9:   # Null type
                return None
        return v

    def walk(node):
        nonlocal status
        
        # Check if this is an operation node: [5, "operationName", ...args]
        if isinstance(node, list) and len(node) >= 2:
            if node[0] == 5 and isinstance(node[1], str):
                op = node[1]
                args = [decode_value(x) for x in node[2:]]
                
                if op == "insertMessageSearchResult":
                    # Based on the structure:
                    # [5, "insertMessageSearchResult", 
                    #   0:query, 1:index, 2:thread_id, 3:?, 4:?, 
                    #   5:sender_name, 6:message_id, 7:timestamp, 
                    #   8:snippet, 9:profile_pic_url, 10:?,
                    #   11:highlight_offset, 12:highlight_length, 13:?]
                    
                    if len(args) >= 13:
                        results.append(MessageSearchResult(
                            query=args[0] or "",
                            result_index=int(args[1]) if args[1] is not None else 0,
                            thread_id=args[2] or "",
                            sender_name=args[5] or "",
                            message_id=args[6] or "",
                            timestamp_ms=int(args[7]) if args[7] else 0,
                            snippet=args[8] or "",
                            profile_pic_url=args[9] or "",
                            highlight_offset=args[11] or "",
                            highlight_length=args[12] or ""
                        ))
                
                elif op == "updateMessageSearchQueryStatus":
                    # [5, "updateMessageSearchQueryStatus",
                    #   0:query, 1:total_results, 2:status, 3:has_more, 4:next_cursor, 5:thread_id, 6:?]
                    if len(args) >= 6:
                        status = MessageSearchStatus(
                            query=args[0] or "",
                            total_results=int(args[1]) if args[1] is not None else 0,
                            status=int(args[2]) if args[2] is not None else 0,
                            has_more=bool(args[3]) if args[3] is not None else False,
                            next_cursor=args[4],
                            thread_id=args[5] or ""
                        )
            # Recursively walk all list elements
            for item in node:
                if isinstance(item, (list, dict)):
                    walk(item)
        
        elif isinstance(node, dict):
            # Also walk dictionary values
            for value in node.values():
                if isinstance(value, (list, dict)):
                    walk(value)

    # Start walking from the root
    walk(inner)
    
    return {"results": results, "status": status}



# Configuration file for the Sphinx documentation builder.
import os
import sys

sys.path.insert(0, os.path.abspath('../../'))

from fbchat_muqit import __version__

project = 'fbchat-muqit'
copyright = '2025, Muhammad MuQiT'
author = 'Muhammad MuQiT'
release = __version__
version = release

# -- General configuration ---------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc.typehints',
    'sphinx.ext.intersphinx',
    # "sphinx.ext.extlinks",
]

# Napoleon settings
napoleon_google_docstring = True 
napoleon_include_init_with_doc = True
napoleon_use_param = True
napoleon_use_rtype = True

# Autodoc settings
autodoc_default_options = {
    'members': True,
    'member-order': 'groupwise',
    'undoc-members': False,
    'show-inheritance': True,
    'inherited-members': True,
    'private-members': False,
    'special-members': False,
}

autodoc_member_order = "bysource"
autodoc_class_signature = "separated"

# Skip submodules that should be excluded from documentation
def skip_submodules(app, what, name, obj, skip, options):
    """Skip members from excluded modules."""

    excluded_modules = [
        'fbchat_muqit.models.deltas',
        'fbchat_muqit.models.mqtt_response',
        'fbchat_muqit.exception',
        'fbchat_muqit.logging',
    ]
    
    if hasattr(obj, '__module__'):
        for excluded in excluded_modules:
            if obj.__module__ and obj.__module__.startswith(excluded):
                return True
    return skip

def setup(app):
    """Sphinx setup hook."""
    app.connect('autodoc-skip-member', skip_submodules)

templates_path = ['_templates']
exclude_patterns = []  # This only affects .rst files, not Python modules

# -- Options for HTML output -------------------------------------------------

html_theme = 'furo'
html_static_path = ['_static']
html_show_sphinx = False
html_show_sourcelink = False

html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/fontawesome.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/solid.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/brands.min.css",
]


html_theme_options = {
    "light_css_variables": {
        "color-brand-primary": "blue",
        "color-brand-content": "#CC3333",
    },
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/togashigreat/fbchat-muqit",
            "html": "",
            "class": "fa-brands fa-solid fa-github fa-2x",
        },
    ],
    "sidebar_hide_name": True,
    "navigation_with_keys": True,
}





from typing import Union
from msgspec import Struct, field 




class friendRequestList(Struct, frozen=True, eq=False, tag="mobile_requests_count", tag_field="type"):
    """New friend request list update. When friend request list is updated"""
    friend_requests: int = field(name="num_unread")
    """current count of friend request"""
    new_friend_request: int = field(name="num_unseen")
    """New friend request count"""


class friendUpdated(Struct, frozen=True, eq=False, tag="jewel_requests_remove_old", tag_field="type"):
    """Removed User id from friend request list either could be because The User cancelled friend request or Client accepted/removed friend request."""
    from_user: int = field(name="from") 
    """Removed User from friend request list"""


class FriendRequestState(Struct, frozen=True, eq=False, tag="friending_state_change", tag_field="type"):
    """Either Friend request confirmed or removed or sent by user"""
    user_id: int = field(name="userid")
    """The action done on the user"""
    action: str
    """action can be 'confirm', 'reject' or 'send'"""


class PokeNotification(Struct, frozen=True, eq=False, tag="live_poke", tag_field="type"):
    """Poked user details"""
    user_poked: int =  field(name="poke_source")
    """The user that Poked Client"""
    poke_time: int
    """The timestamp of the Poke"""



class PageNotification(Struct, frozen=True, eq=False):
    """Page notifucation extra information"""
    sender_id: str = field(name="senderId")
    """The `User` who sent message to Client's `Page`"""
    page_id: str = field(name="pageId")
    """The Id of the Page"""
    page_name: str = field(name="pageName")
    """Page's name"""
    message_id: str = field(name="messageId")
    """The Id of the sent message"""
    title: str 
    """Title of the notifucation"""
    text: str = field(name="body")
    """Text of the notifucation. Usually the text message that was sent to Page."""
    sender_profile_pic: str = field(name="senderProfPicUrl")
    """The user's profile picture"""
    page_profile_pic: str = field(name="pageProfPicUrl")
    """The Page's profile picture"""




class friendRequestSeen(Struct, frozen=True, eq=False, tag="friend_requests_seen", tag_field="type"):
    pass


class NotificationSeen(Struct, frozen=True, eq=False, tag="notifications_seen", tag_field="type"):
    pass

class NavigationNotify(Struct, frozen=True, eq=False, tag="nav_update_counts", tag_field="type"):
    pass

class ConfirmedFriend(Struct, frozen=True, eq=False, tag="jewel_friending_notifs", tag_field="type"):
    pass


FacebookNotifications = Union[friendRequestList, friendUpdated, FriendRequestState, PokeNotification, friendRequestSeen, NotificationSeen,NavigationNotify, ConfirmedFriend]



"""
Instead of dataclasses or attrs, msgspec's `Struct` classes are used to automatically decode and Parse received json payloads. 
"""

from __future__ import annotations



from typing import List, Optional, Union
from msgspec import Struct, field

from .custom_type import (
        DecodedPayloadType, 
        MentionType, 
    )

from .attachments_deltas import (
        RawAttachments,
        )

# Delta classss
from .group_deltas import MessengerGroupDeltas
from ..timestamps import TimestampDeltas

from ..messagesData import MessageData
from .payloads import DeltaMessageReply
from ..typing import Typing 
from ..presence import Presence


class ClientPayloadDelta(Struct, frozen=True, eq=False):
    """"""
    deltas: List[DeltaMessageReply]

class ClientPayload(Struct, frozen=True, eq=False, tag="ClientPayload", tag_field="class"):
    payload: DecodedPayloadType  # receives `ClientPayloadDelta` object


class NewMessageDelta(Struct, frozen=True, eq=False, tag="NewMessage", tag_field="class"):

    messageMetadata: MessageData
    body: Optional[str] = None
    attachments: List[RawAttachments] = list()
    mentions: MentionType = field(name="data", default=MentionType())
    participants: tuple = tuple()


class NoOpDelta(Struct, frozen=True, eq=False, tag="NoOp", tag_field="class"):
    """"""

DeltaTypes = Union[NewMessageDelta, ClientPayload, NoOpDelta, TimestampDeltas, MessengerGroupDeltas]

# ------------------ Main Parent Delta wrappers ------------------- #

class DeltaWrapper(Struct, frozen=True, eq=False):
    deltas: List[DeltaTypes]  # automatically parsed by msgspec based on tag
    firstDeltaSeqId: Optional[int] = None
    lastIssuedSeqId: Optional[int] = None
    syncToken: Optional[str] = None

# class Typing is Parent wrapper

# class Presence is Parent Wrapper 


class FirstFetch(Struct, frozen=True, eq=False):
    firstDeltaSeqId: int
    queueEntityId: int
    syncToken: str


# ------------------ End Delta wrappers ------------------- #




"""
For Type annotation you can import all models from here.
"""

from .thread import ThreadType, ThreadFolder, Thread
from .user import User

from .attachment import (
        Attachment,
        AttachmentType,
        ImageAttachment,
        VideoAttachment,
        GifAttachment,
        StickerAttachment,
        AudioAttachment,
        FileAttachment,

        LocationAttachment,
        PostAttachment,
        SharedAttachment,
        ReelAttachment,
        ProfileAttachment,
        ProductAttachment,
        ExternalAttachment,

        Post,
        Author,
        GroupInfo,
        Media,
        Image,
        Dimension
        )

from .thread_actions import (
        ApprovalMode,
        ApprovalQueue,
        AdminAdded,
        AdminRemoved,
        ApprovedUser,
        addedParticipant,
        ChangeViwerStatus,
        JoinableMode,
        ParticipantsAdded,
        ParticipantLeft,
        ThreadAction,
        ThreadDelete,
        ThreadEmoji,
        ThreadFolderMove,
        ThreadMagicWord,
        ThreadMessagePin,
        ThreadMessageUnPin,
        ThreadMessageSharing,
        ThreadMuteSettings,
        ThreadName,
        ThreadNickname,
        ThreadTheme,
        MuteThread,
        ForcedFetch
    )

from .typing import Typing

from .timestamps import ReadReceipt, DeliveryReceipt, MarkFolderSeen, MarkRead, MarkUnread

from .message import Message, Mention, Mentions, MessageType, MessageRemove, MessageReaction, MessageUnsend, EmojiSize, Reaction

from .messagesData import MessageData

from .notifications import PokeNotification, FriendRequestState, friendUpdated, friendRequestList, PageNotification

from .presence import Presence, UserStatus

from .themes import Theme, AlternativeTheme, Asset

from .mqtt_response.search_message import MessageSearchResult, MessageSearchStatus



__all__ = [
    # Thread related
    "ThreadType", "ThreadFolder", "Thread",
    
    # User
    "User",
    
    # Attachments
    "Attachment",
    "AttachmentType",
    "ImageAttachment",
    "VideoAttachment",
    "GifAttachment",
    "StickerAttachment",
    "AudioAttachment",
    "FileAttachment",
    "LocationAttachment",
    "PostAttachment",
    "SharedAttachment",
    "ReelAttachment",
    "ProfileAttachment",
    "ProductAttachment",
    "ExternalAttachment",
    "Post",
    "Author",
    "GroupInfo",
    "Media",
    "Image",
    "Dimension",
    
    # Thread Actions
    "ApprovalMode",
    "ApprovalQueue",
    "AdminAdded",
    "AdminRemoved",
    "ApprovedUser",
    "addedParticipant",
    "ChangeViwerStatus",
    "JoinableMode",
    "ParticipantsAdded",
    "ParticipantLeft",
    "ThreadAction",
    "ThreadDelete",
    "ThreadEmoji",
    "ThreadFolderMove",
    "ThreadMagicWord",
    "ThreadMessagePin",
    "ThreadMessageUnPin",
    "ThreadMessageSharing",
    "ThreadMuteSettings",
    "ThreadName",
    "ThreadNickname",
    "ThreadTheme",
    "MuteThread",
    "ForcedFetch",
    
    # Typing
    "Typing",
    
    # Timestamps
    "ReadReceipt",
    "DeliveryReceipt",
    "MarkFolderSeen",
    "MarkRead",
    "MarkUnread",
    
    # Messages
    "Message",
    "Mention",
    "Mentions",
    "MessageType",
    "MessageRemove",
    "MessageReaction",
    "MessageUnsend",
    "Reaction",
    "EmojiSize",
    
    # Message Data
    "MessageData",
    
    # Notifications
    "PokeNotification",
    "PageNotification",
    "FriendRequestState",
    "friendUpdated",
    "friendRequestList",
    
    # Presence
    "Presence",
    "UserStatus",
    
    # Themes
    "Theme",
    "AlternativeTheme",
    "Asset",

    # Message Response 
    "MessageSearchStatus",
    "MessageSearchResult",
]

__title__ = "fbchat-muqit"
__version__ = "1.1.5"
__description__ = "A powerful Facebook Messenger Chat API to interact with Facebook and Messenger. Easy to use fast and efficient library"

__copyright__ = "Copyright 2024 by Muhammad MuQiT"

__license__ = "GPL-V3.0"

__author__ = "Muhammad MuQiT"
__email__ = "togashiyuuta1111@gmail.com"




#fbchat_muqit/facbook/client.py

import asyncio
import json
import base64
import aiofiles
from pathlib import Path
from enum import Enum

from typing import List, Optional

from msgspec.json import Decoder


from ..state import State
from ..muqit import Mqtt
from ..logging.logger import FBChatLogger, get_logger 
from ..exception.errors import APIError, LoginError, ParsingError, ValidationError
from ..models.message import Mention
from .response_parser import PictureUploadResponse, Privacy, Audience, OverridenPrivacy, PrivacyRow,ResponsePostData, extract_privacy_data

from ..utils.utils import now, generate_uuid
from ..models.deltas.parser import MessageParser

class FBReaction(Enum):
    LIKE = "1635855486666999"
    LOVE = "1678524932434102"
    CARE = "613557422527858"
    HAHA = "115940658764963"
    WOW = "478547315650144"
    SAD = "908563459236466"
    ANGRY = "444813342392137"




def mention_to_dict(mentions: List[Mention]):
    return [
            {"entity": {"id": m.user_id }, "length": m.length, "offset": m.offset}
            for m in mentions
        ]



def post_attachments(picture_ids: Optional[List[str]] = None, video_ids: Optional[List[str]] = None): 
    if picture_ids and video_ids:
        return [{"photo": {"id": i}} for i in picture_ids] + [{ "video": {"audio_descriptions": None, "id": i, "notify_when_processed": True, "transcriptions": None, "was_created_via_unified_video_flow": None} } for i in video_ids]

    elif picture_ids:
        return [{"photo": {"id": i}} for i in picture_ids]
    elif video_ids:
        return [{ "video": {"audio_descriptions": None, "id": i, "notify_when_processed": True, "transcriptions": None, "was_created_via_unified_video_flow": None} } for i in video_ids]
    else:
        return []




class FacebookClient:
    """`FacebookClient` handles facebook posts, reaction, friend request etc. and for publishing post, commenting and reacting to Facebook posts."""
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._state: Optional[State] = None
        self._uid: str = ""
        self._name: str = ""
        self._parser = MessageParser(self.logger)
        self.logger: FBChatLogger = get_logger()
        self._mqtt: Optional[Mqtt] = None
        self._client_mutation_id = 0
        # response parser 
        self._privacyResponse_decoder = Decoder(type=OverridenPrivacy, strict=False)
        self._picture_upload_dec = Decoder(type=PictureUploadResponse, strict=False)
        self._post_create_dec = Decoder(type=ResponsePostData, strict=False) 
        # needed in headers for video upload
        self._origin = "https://www.facebook.com"
        self._referer = "https://www.facebook.com/"
        if self._state:
            self._origin = f"https://{self._state._host}"
            self._referer = f"https://{self._state._host}/"


    def get_mutation_id(self)-> str:
        m = self._client_mutation_id
        self._client_mutation_id += 1 
        return str(m)


    async def manage_friend_request(self, user_id: str, accept_request: bool = True):
        """Manage friend requests accept or reject a friend requests

        Args:
            user_id (str): The user's id you want to accept or reject the friend requests.
            accept_request (bool): Accepts the friend request if true else rejects it.
        """
        if not self._state:
            raise LoginError("Client is not logged in yet. `State` class is not initialised yet")
        ts = str(now())
        variables = {
            "input": {
                "click_correlation_id": ts,
                "click_proof_validation_result": "{\"validated\":true}",
                "friend_requester_id": user_id,   # The friend request sender
                "friending_channel": "FRIENDS_HOME_REQUESTS",
                "actor_id": self._uid,   # <-- your own ID (Client)
                "client_mutation_id": self.get_mutation_id(),  
                },
                "scale": 3,
                "refresh_num": 0,
            } 

        data = {
            "fb_api_caller_class": "RelayModern",
            "fb_api_req_friendly_name": "FriendingCometFriendRequestDeleteMutation",
            "server_timestamps": True,
            "doc_id":  "25003074442651692"
            }

        if accept_request:
            data["fb_api_req_friendly_name"] =  "FriendingCometFriendRequestConfirmMutation"
            data["doc_id"] =  "24205795295769853"
            variables["input"]["warn_ack"] = False
            variables["should_fix_banner"] = True 

        data["variables"] = json.dumps(variables)

        await self._state._post("https://www.facebook.com/api/graphql/", data=data, no_response=True)


    async def send_friend_request(self, user_ids: List[str]):
        """
        Send friend request to one or multiple users using their Id.

        Args:
            user_ids (List[str]): A list of user ids to send friend request.
        """
        data = {
            "fb_api_caller_class": "RelayModern",
            "fb_api_req_friendly_name": "FriendingCometFriendRequestSendMutation",
            "server_timestamps": True,  # Converted to proper boolean
            "variables": json.dumps({
                "input": {
                    "click_correlation_id": str(now()),
                    "click_proof_validation_result": "{\"validated\":true}",
                    "friend_requestee_ids": user_ids,
                    "friending_channel": "FRIENDS_HOME_MAIN",
                    "warn_ack_for_ids": [],
                    "actor_id": self._uid,
                    "client_mutation_id": self.get_mutation_id()
                    },
                "scale": 3
                }),
            "doc_id": "24974393785534352"
            }   
        if self._state:
            await self._state._post("/api/graphql/", data=data, no_response=True)
    
    async def unfriend(self, user_id: str):
        """Unfriend a friend using their Id.
        
        Args:
            user_id (str): Id of the User to unfriend.
        """
        
        data = {
                "fb_api_caller_class": "RelayModern",
                "fb_api_req_friendly_name": "FriendingCometUnfriendMutation",
                "server_timestamps": True,
                "variables": json.dumps({
                    "input": {
                        "source": "bd_profile_button",
                        "unfriended_user_id": user_id,
                        "actor_id": self._uid,
                        "client_mutation_id": self.get_mutation_id()
                        },
                    "scale": 3
                    }),
                "doc_id": "24028849793460009"
                }
        if self._state:
            await self._state._post("/api/graphql/", data=data, raw=True)

    async def cancel_friend_request(self, user_id: str):
        """Cancel a friend request that you sent using the requested to user Id.

        Args:
            user_id (str): The Id of the User you sent friend request to.
        """
        data = {
            "fb_api_caller_class": "RelayModern",
            "fb_api_req_friendly_name": "FriendingCometFriendRequestCancelMutation",
            "server_timestamps": True,
            "variables": json.dumps({
                "input": {
                    "cancelled_friend_requestee_id": user_id,
                    "click_correlation_id": str(now()),
                    "click_proof_validation_result": "{\"validated\":true}",
                    "friending_channel": "PROFILE_BUTTON",
                    "actor_id": self._uid,
                    "client_mutation_id": self.get_mutation_id()
                    },
                "scale": 3
                }),
            "doc_id": "24453541284254355"
            }

        if self._state:
            await self._state._post("/api/graphql/", data=data, raw=True)



    async def react_to_post(self, feedback_id: Optional[str] = None , post_id: Optional[int] = None, reaction: FBReaction = FBReaction.LOVE):
        """React to a post using post's feedback Id or Post Id

        Args:
            feedback_id (Optional[str]): The feedback Id of the post to react. 
            post_id (Optional[int]): The Id of post to react. (feedback_id or post_id one of them must be provided)
            reaction (FBReaction): The reaction to react with (LOVE, HAHA, SAD e.g.)
        """
        if not self._state:
            raise LoginError("Client is not logged in yet. `State` class is not initialised yet")
        if not feedback_id and not post_id:
            raise ValidationError("Either 'feedback_id' or 'post_id' must be provided to react to Facebook post.")

        if post_id:
            feedback_id = base64.b64encode(post_id.to_bytes((post_id.bit_length() + 7) // 8, byteorder='big')).decode('utf-8')

        data = {
            "lsd": self._state._lsd,
            "fb_api_caller_class": "RelayModern",
            "fb_api_req_friendly_name": "CometUFIFeedbackReactMutation",
            "server_timestamps": True,
            "variables": json.dumps({
                "input": {
                    "attribution_id_v2": f"CometHomeRoot.react,comet.home,tap_tabbar,{now()},420553,4748854339,,",
                    "feedback_id": feedback_id,
                    "feedback_reaction_id": reaction.value,
                    "feedback_source": "NEWS_FEED",
                    "is_tracking_encrypted": True,
                    "tracking": [],  # intentionally left empty
                    "session_id": generate_uuid(),
                    "actor_id": self._uid,
                    "client_mutation_id": self.get_mutation_id()
                    },
            "useDefaultActor": False,
            "__relay_internal__pv__CometUFIReactionsEnableShortNamerelayprovider": False,
            }),
            "doc_id": "24034997962776771"
        }

        await self._state._post("/api/graphql/", data=data, no_response=True)

    async def _pick_container_query(self, privacy_writer_id, privacy = None):
        data = {
            "fb_api_caller_class": "RelayModern",
            "fb_api_req_friendly_name": "CometPrivacySelectorPickerContainerQuery",
            "server_timestamps": True,
            "variables": json.dumps({
                "localPrivacyRow": privacy,
                "privacyWriteID": privacy_writer_id,  # long encoded string kept as-is
                "renderLocation": "COMET_COMPOSER",
                "scale": 3
                }),
            "doc_id": "24820345800985339",
        }
        if self._state:
            await self._state._post("/api/graphql/", data=data, no_response=True)



    async def _get_privacy_writer(self):
        if not self._state:
            raise LoginError("Client is not logged in yet. `State` class is not initialised yet")
        data = {
                "fb_api_caller_class": "RelayModern",
                "fb_api_req_friendly_name": "FeedComposerCometRootQuery",
                "server_timestamps": True,
                "variables": json.dumps({
                    "hasStory": False,
                    "isBizWeb": False,
                    "privacySelectorRenderLocation": "COMET_COMPOSER",
                    "profileID": self._uid,
                    "scale": 3,
                    "storyID": "",
                    "__relay_internal__pv__FeedComposerComet_isGenAILabelEnabledrelayprovider": False,
                    "__relay_internal__pv__CometUnifiedVideoCreation_showPrivacyMergereLayprovider": False,
                    "__relay_internal__pv__CometUGCPublicCreation_showComposerPublicAwarenessTooltiprelayprovider": False
                    }),
                "doc_id": "32319398104317803"                
            }
            
        r = await self._state._post("/api/graphql/", data=data, raw=True)
        # self.logger.info(f"{r[:700]}")
        # with open("privacy_writer.txt", "w") as f:
        #     f.write(r.decode())
        # may show errors in some accounts
        try:
            r = extract_privacy_data(r.decode())
        except Exception as e:
            raise ParsingError("Failed to parse graphql response while getting privacy writer Id to set audience type.", original_exception=e)
        return r



    async def _setPrivacyWriter(self, ids: List, privacyWriter: PrivacyRow, isAllow = False, isDeny = False, base_state = None)-> Privacy | None:
        if not self._state:
            raise LoginError("Client is not logged in yet. `State` class is not initialised yet")
        data = {
                "lsd": self._state._lsd,
                "fb_api_caller_class": "RelayModern",
                "fb_api_req_friendly_name": "refetchCometPrivacySelectorNonAutosavePickerQuery",
                "server_timestamps": True,
                "variables": json.dumps({
                    "localPrivacyRow": { 
                        "allow": ids if isAllow else [],
                        "base_state": base_state if base_state else privacyWriter.privacy_row_input.base_state.value,
                        "deny": ids if isDeny else [],
                        "tag_expansion_state":"UNSPECIFIED"
                        },
                    "privacyWriteID": privacyWriter.id,
                    "renderLocation":"COMET_COMPOSER",
                    "scale":3,
                    "tags": None
                    }),
                "doc_id": "24578653808469895",
                }

         
        r = await self._state._post("/api/graphql/", data=data, raw=True)
        if base_state:
            return
        r = self._privacyResponse_decoder.decode(r)
        return r.data.node.scope.selected_row_override


    async def _format_audience(self, specific_users, except_users, base_state):
        if specific_users and except_users:
            raise APIError("Provide 'specific_users' or 'except_users' only one at a time.")

        privacy_writer = await self._get_privacy_writer()
        await self._pick_container_query(privacy_writer.id)

        if base_state != privacy_writer.privacy_row_input.base_state.value and not specific_users and not except_users:
            await self._setPrivacyWriter([], privacy_writer, base_state=base_state)

        if specific_users:
            data = await self._setPrivacyWriter(specific_users, privacy_writer, isAllow=True)
            if data:
                specific_users = data.specific_users
                except_users = data.except_users 
                base_state = data.base_state.value

        elif except_users:
            data = await self._setPrivacyWriter(except_users, privacy_writer, isDeny=True)
            if data:
                specific_users = data.specific_users
                except_users = data.except_users 
                base_state = data.base_state.value

        form =  {
            "privacy": {
                "allow": specific_users or [],
                "base_state": base_state,
                "deny": except_users or [],
                "tag_expansion_state": "UNSPECIFIED",
                }
            }
        return form


    async def publish_post(self, text: str, image_paths: Optional[List[str]] = None, video_paths: Optional[List[str]] = None, tag_users: Optional[List[str]] = None,  specific_users: Optional[List[str]] = None, except_users: Optional[List[str]] = None, audience: Audience = Audience.PUBLIC, mentions: Optional[List[Mention]] = None) -> str:
        """Publish a post in facebook.

        Args:
            text (str): The text message you want to add in the post. 
            image_paths (Optional[str]): The Images you want to attach to the post. 
            video_paths (Optional[List[str]]): The video to attach to the post. 
            tag_users (Optional[str]): Tag users to the postu using their uid. 
            specific_users (Optional[str]): only show the post to specific users. 
            except_users (Optional[str]): Show the post except some users using their uid. 
            audience (Audience): The audience of the post can be FRIENDS, EVERYONE, ONLYME.
            mentions (Optional[List[Mention]]): The users you want to mention in the post. 

        Returns:
            str: If successfully published the post returns post's feedback Id (base64 encoded post id) or post Id.
        """
        if not self._state:
            raise LoginError("Client is not logged in yet. `State` class is not initialised yet")
        
        # Build attachments list
        attachments = []
        
        # Upload files if paths are provided
        uploaded_photo_ids = None
        uploaded_video_ids = None
        
        if image_paths:
            self.logger.info(f"Uploading {len(image_paths)} photos...")
            uploaded_photo_ids = await self.upload_photos(image_paths)
        
        if video_paths:
            self.logger.warning(f"Due to some issues attaching video to post is not supported at this moment. Skipping videos.")
            # self.logger.info(f"Uploading {len(video_paths)} videos...")
            # uploaded_video_ids = await self.upload_videos(video_paths)
        
        
        # Build attachments
        if uploaded_photo_ids or uploaded_video_ids:
            attachments = post_attachments(
                picture_ids=uploaded_photo_ids,
                video_ids=uploaded_video_ids
            )

        self.logger.info(f"attachments: {attachments}")
        self.logger.info("Now getting audience...")
        audience_data =  await self._format_audience(specific_users, except_users, audience.value)
        new_uuid = generate_uuid()
        variables = {
            "input": {
            "composer_entry_point": "inline_composer",
            "composer_source_surface": "newsfeed",
            "composer_type": "feed",
            "idempotence_token": f"{new_uuid}_FEED",
            "source": "WWW",
            "audience": audience_data,
            "message": {
                "ranges": mention_to_dict(mentions) if mentions else [],
                "text": text,
                },
            "inline_activities": [],
            "text_format_preset_id": "0",
            "publishing_flow": {
                "supported_flows": ["ASYNC_SILENT", "ASYNC_NOTIF", "FALLBACK"],
                },
            "attachments": attachments,
            "with_tags_ids": tag_users,
            "logging": {
                "composer_session_id": new_uuid,
                },
            "navigation_data": {
                "attribution_id_v2": f"CometHomeRoot.react,comet.home,via_cold_start,{now()},929297,4748854339,,"
                },
            "tracking": [None],
            "event_share_metadata": {
                "surface": "newsfeed",
                },
            "actor_id": self._uid,
            "client_mutation_id": "1",
            },
            "feedLocation": "NEWSFEED",
            "feedbackSource": 1,
            "focusCommentID": None,
            "gridMediaWidth": None,
            "groupID": None,
            "scale": 3,
            "privacySelectorRenderLocation": "COMET_STREAM",
            "checkPhotosToReelsUpsellEligibility": True,
            "renderLocation": "homepage_stream",
            "useDefaultActor": False,
            "inviteShortLinkKey": None,
            "isFeed": True,
            "isFundraiser": False,
            "isFunFactPost": False,
            "isGroup": False,
            "isEvent": False,
            "isTimeline": False,
            "isSocialLearning": False,
            "isPageNewsFeed": False,
            "isProfileReviews": False,
            "isWorkSharedDraft": False,
            "hashtag": None,
            "canUserManageOffers": False,
            "__relay_internal__pv__CometUFIShareActionMigrationrelayprovider": True,
            "__relay_internal__pv__GHLShouldChangeSponsoredDataFieldNamerelayprovider": True,
            "__relay_internal__pv__GHLShouldChangeAdIdFieldNamerelayprovider": True,
            "__relay_internal__pv__CometUFI_dedicated_comment_routable_dialog_gkrelayprovider": False,
            "__relay_internal__pv__CometUFICommentAvatarStickerAnimatedImagerelayprovider": False,
            "__relay_internal__pv__IsWorkUserrelayprovider": False,
            "__relay_internal__pv__CometUFIReactionsEnableShortNamerelayprovider": False,
            "__relay_internal__pv__FBReels_enable_view_dubbed_audio_type_gkrelayprovider": True,
            "__relay_internal__pv__FBReels_deprecate_short_form_video_context_gkrelayprovider": True,
            "__relay_internal__pv__FeedDeepDiveTopicPillThreadViewEnabledrelayprovider": False,
            "__relay_internal__pv__CometImmersivePhotoCanUserDisable3DMotionrelayprovider": False,
            "__relay_internal__pv__WorkCometIsEmployeeGKProviderrelayprovider": False,
            "__relay_internal__pv__IsMergQAPollsrelayprovider": False,
            "__relay_internal__pv__FBReels_enable_meta_ai_label_gkrelayprovider": True,
            "__relay_internal__pv__FBReelsMediaFooter_comet_enable_reels_ads_gkrelayprovider": True,
            "__relay_internal__pv__StoriesArmadilloReplyEnabledrelayprovider": True,
            "__relay_internal__pv__FBReelsIFUTileContent_reelsIFUPlayOnHoverrelayprovider": True,
            "__relay_internal__pv__GroupsCometGYSJFeedItemHeightrelayprovider": 150,
            "__relay_internal__pv__StoriesShouldIncludeFbNotesrelayprovider": False,
            "__relay_internal__pv__GHLShouldChangeSponsoredAuctionDistanceFieldNamerelayprovider": False,
            "__relay_internal__pv__GHLShouldUseSponsoredAuctionLabelFieldNameV1relayprovider": False,
            "__relay_internal__pv__GHLShouldUseSponsoredAuctionLabelFieldNameV2relayprovider": False,
        }

        data = {
            "lsd": self._state._lsd,
            "fb_api_caller_class": "RelayModern",
            "fb_api_req_friendly_name": "ComposerStoryCreateMutation",
            "server_timestamps": True,
            "variables": json.dumps(variables),  # This gets JSON stringified
            # "doc_id": "9137564299700449", 
            # "doc_id": "24309936402019358"
            # multiple doc ids were found 
            "doc_id": "24966185093062904"
            }
        headers = {
                "Sec-Fetch-Mode": "cors",
                "Sec-Fetch-Site": "same-origin",
                "Origin": self._origin,
                "Referer":  self._referer
                }
        r = await self._state._post("/api/graphql/", data=data, raw=True, header_type="publish_post", headers=headers)
        try:
            r = self._post_create_dec.decode(r)
        except Exception as e:
            raise ParsingError("Failed to parse graphql response of post publishing. Coudn't get `feedback_id` or `post_id`", original_exception=e)
        if r.data.story_create.story_id:
            return r.data.story_create.story_id
        elif r.data.story_create.story:
            return r.data.story_create.story.id
        else: 
            return str(r.data.story_create.post_id)




    async def upload_photo(self, image_path: str)->str:
        """
        Upload a photo to Facebook and return the photo ID.
    
        Args:
            image_path (str): Path to the image file to upload
        
        Returns:
            str : Upload response containing photoID
        """
        if not self._state:
            raise LoginError("Client is not logged in yet. `State` class is not initialised yet")
        
        # Generate upload ID
        upload_id = f"jsc_c_{generate_uuid()[:8]}"
    
        # Read file
        file_path = Path(image_path)
        if not file_path.exists():
            raise ValidationError(f"File not found: {image_path}")
    
        async with aiofiles.open(file_path, 'rb') as f:
            file_data = await f.read()

        data = {
                "lsd": self._state._lsd,
                "source": "8",
                "profile_id": self._uid,
                "waterfallxapp": "comet",
                "upload_id": upload_id
                }
        files = {"farr": (file_path.name, file_data, self._state._magic.from_buffer(file_data))}
    
        # Upload URL from your screenshot
        url = "https://upload.facebook.com/ajax/react_composer/attachments/photo/upload"
        headers = {
                "Sec-Fetch-Mode": "cors",
                "Sec-Fetch-Site": "same-site",
                }
        
        response = await self._state._post(url, data=data, files=files, raw=True, headers=headers)
        response = self._picture_upload_dec.decode(response[response.index(b'{'):])
        return response.payload.photoID
        
        
    async def upload_photos(self, image_paths: List[str], max_concurrent: int = 5)->List[str]:
        """
        Upload multiple photos concurrently and return list of photo IDs.
        
        Args:
            image_paths (List[str]): List of image file paths to upload
            max_concurrent (int): Maximum number of concurrent uploads (default: 5)
            
        Returns:
            List[str]: List of successfully uploaded photo IDs (preserves order)
        """

        semaphore = asyncio.Semaphore(max_concurrent)
        async def upload_with_semaphore(path: str, index: int)-> str:
            """Upload with semaphore and return (index, photo_id)"""
            async with semaphore:
                self.logger.info(f"Uploading {index + 1}/{len(image_paths)}: {path}")
                photo_id = await self.upload_photo(path)
                return photo_id

        # Create tasks for all uploads
        tasks = [
            upload_with_semaphore(path, idx) 
            for idx, path in enumerate(image_paths)
        ]
        # Wait for all uploads to complete
        results = await asyncio.gather(*tasks, return_exceptions=True)

        photo_ids = []
        errors = []
        
        for result in results:
            if isinstance(result, Exception):
                self.logger.error(f"Upload task failed with exception: {result}")
                errors.append(result)
            elif result is not None:  # (index, photo_id)
                photo_ids.append(result)
            else:
                errors.append(f"Upload failed for index {result[0]}")
        
        # Sort by original index to maintain order
     
        self.logger.info(f"Upload complete: {len(photo_ids)}/{len(image_paths)} successful")
        
        if errors:
            self.logger.warning(f"Encountered {len(errors)} errors during upload")
        
        return photo_ids

    # # ================= Video =============== s 
    #
    #
    # async def _get_video_upload_config(self) -> dict:
    #     """Get video uploader configuration."""
    #     if not self._state:
    #         raise LoginError("Client is not logged in yet.")
    #
    #     data = {
    #         "fb_api_caller_class": "RelayModern",
    #         "fb_api_req_friendly_name": "useCometVideoUploaderConfigQuery",
    #         "server_timestamps": True,
    #         "variables": json.dumps({
    #             "actorID": self._uid,
    #             "entryPoint": "feed",
    #             "targetID": ""
    #         }),
    #         "doc_id": "9734072893355148"
    #     }
    #
    #     response = await self._state._post(
    #         "/api/graphql/",
    #         data=data,
    #         raw=True
    #     )
    #
    #     data = json.loads(response.decode())
    #     config_str = data.get("data", {}).get("viewer", {}).get("comet_composer_video_uploader_config", "{}")
    #     config = json.loads(config_str)
    #
    #     return config
    #
    # async def _get_media_upload_config(self) -> dict:
    #     """Get media upload configuration."""
    #     if not self._state:
    #         raise LoginError("Client is not logged in yet.")
    #
    #
    #     data = {
    #         "fb_api_caller_class": "RelayModern",
    #         "fb_api_req_friendly_name": "MediaUploadFBDefaultServerConfigurationRetrieverQuery",
    #         "server_timestamps": True,
    #         "variables": json.dumps({
    #             "source_type": "newsfeed_composer"
    #         }),
    #         "doc_id": "24229633186643574"
    #     }
    #
    #     response = await self._state._post("/api/graphql/", data=data, raw=True) 
    #
    #     media = json.loads(response.decode())
    #     media = media["data"]["media_upload_config"]
    #     return {
    #             "start_uri": media["network_start"]["uri"],
    #             "receive_uri": media["network_receive"]["uri"]
    #             }
    #
    #
    #
    # async def _start_video_upload(self, file_size: int, config: dict) -> dict:
    #     """Start video upload session and get video_id."""
    #     if not self._state:
    #         raise LoginError("Client is not logged in yet.")
    #
    #     waterfall_id = generate_uuid().replace("-", "")
    #
    #     # the form data for start request
    #     form_data = {
    #         "waterfall_id": waterfall_id,
    #         "target_id": self._uid,
    #         "source": "newsfeed_composer",
    #         "composer_entry_point_ref": "feed",
    #         "supports_chunking": "true",
    #         "supports_file_api": "true",
    #         "file_size": str(file_size),
    #         "file_extension": "mp4",
    #         "partition_start_offset": "0",
    #         "partition_end_offset": str(file_size),
    #         "composer_dialog_version": "V2",
    #         "video_publisher_action_source": "",
    #         "lsd": self._state._lsd,
    #     }
    #     headers = {
    #             "Sec-Fetch-Mode": "cors",
    #             "Sec-Fetch-Site": "same-origin",
    #             "X-Fb-Lsd": self._state._lsd,
    #             "X-fb-video-waterfall_id": waterfall_id,
    #             }
    #
    #     # Start URL from config
    #     start_url = config.get("start_uri", f"https://{self._state._host}/ajax/video/upload/requests/start/")
    #
    #     response = await self._state._post(start_url + f"?av={self._uid}&__a=1", data=form_data, raw=True, headers=headers)
    #
    #     # Parse response
    #     # remove "for (;;);" by 9: to parse to json
    #     response = self._parser.decoder.decode(response[9:])
    #
    #     payload = response.get("payload", {})
    #     return {
    #         "video_id": payload.get("video_id"),
    #         "start_offset": payload.get("start_offset", 0),
    #         "end_offset": payload.get("end_offset", file_size),
    #         "skip_upload": payload.get("skip_upload", False),
    #         "waterfall_id": waterfall_id
    #     }
    #
    #
    #
    # async def _prepare_for_upload_chunk(self, upload_url):
    #     if not self._state:
    #         raise LoginError("Client is not logged in yet.")
    #     data = { "fb_dtsg_ag": self._state._fb_dtsg_ag, "jazoest": self._state._jazoest_async }
    #     await self._state._get(upload_url, params=data)
    #
    #
    # async def _get_request_access(self, upload_url, params):
    #     if not self._state:
    #         raise LoginError("Client is not logged in yet.")
    #     headers = {
    #             "Access-Control-Request-Headers": "composer_session_id,end_offset,id,offset,product_media_id,start_offset,x-entity-length,x-entity-name,x-entity-type,x-total-asset-size",
    #             "Access-Control-Request-Method": "POST",
    #             "Sec-Fetch-Dest": "empty",
    #             "Sec-Fetch-Mode": "cors",
    #             "Sec-Fetch-Site": "same-site",
    #             "Origin": self._origin,
    #             "Referer": self._referer
    #             }
    #
    #     await self._state._option(upload_url, params, headers)
    #
    #
    # async def _upload_video_chunk(
    #     self, 
    #     upload_url: str,
    #     video_id: str,
    #     video_data: Tuple[str, bytes, str],
    #     waterfall_id: str,
    #     start_offset: str,
    #     end_offset: str,
    #     video_size: str
    # ) -> str:
    #     if not self._state:
    #         raise LoginError("Client is not logged in yet.")
    #
    #     # From response 5: {"h":"1:MTAwMDA3MDgyMS5tcDQ=:video/mp4:..."}
    #     # Prepare headers for chunk upload
    #     fileData = {"": video_data}
    #     data = {"lsd": self._state._lsd, "__aaid": "0"}
    #     headers = {
    #             "Content-Length": video_size,
    #             "Composer_session_id": waterfall_id,
    #             "Product_media_id": video_id,
    #             "End_offset": str(end_offset),
    #             "Offset": str(start_offset),
    #             "Start_offset": str(start_offset),
    #             "Id": "undefined",
    #             "X-Entity-Length": video_size,
    #             "X-Entity-Name": "1000072022.mp4",
    #             "X-Entity-Type": "video/mp4",
    #             "X-Total-Asset-Size": video_size,
    #             "Origin": self._origin,
    #             "Referer":  self._referer,
    #             "Sec-Fetch-Mode": "cors",
    #             "Sec-Fetch-Site": "same-site",
    #             }
    #
    #     await self._get_request_access(upload_url, data)
    #
    #     response = await self._state._post(upload_url, data={}, params=data, files=fileData, raw=True, headers=headers) #type: ignore
    #     result = json.loads(response.decode())
    #     self.logger.debug(f"Received Upload Handler: {result}")
    #     upload_handle = result.get("h")  # This is the chunk upload handle 
    #
    #     return upload_handle
    #
    #
    # async def _finalize_video_upload(
    #     self,
    #     video_id: str,
    #     start_offset: str, 
    #     end_offset: str, 
    #     upload_handle: str,
    #     waterfall_id: str,
    #     config: dict,
    #     file_size: int
    # ) -> dict:
    #     """Finalize video upload after chunks are uploaded."""
    #     if not self._state:
    #         raise LoginError("Client is not logged in yet.")
    #
    #     # receive endpoint
    #     form_data = {
    #         "waterfall_id": waterfall_id,
    #         "target_id": self._uid,
    #         "video_id": video_id,
    #         "source": "newsfeed_composer",
    #         "composer_entry_point_ref": "feed",
    #         "supports_chunking": "true",
    #         "supports_upload_service": "true",
    #         "partition_start_offset": str(start_offset),
    #         "partition_end_offset": str(end_offset),
    #         "start_offset": str(start_offset),
    #         "end_offset": str(end_offset),
    #         "upload_speed": "741605.047318612",  # Calculate based on actual speed
    #         "fbuploader_video_file_chunk": upload_handle,
    #         "composer_dialog_version": "V2",
    #         "lsd": self._state._lsd,
    #     }
    #     headers = {
    #             "X-Fb-Lsd": self._state._lsd,
    #             "X-fb-video-waterfall_id": waterfall_id,
    #             "Sec-Fetch-Mode": "cors",
    #             "Sec-Fetch-Site": "same-origin",
    #             }
    #
    #     receive_url = config.get("receive_uri", f"https://{self._state._host}/ajax/video/upload/requests/receive/")
    #
    #     response = await self._state._post(receive_url + f"?av={self._uid}&__a=1", data=form_data, raw=True, headers=headers)
    #
    #     # remove "for (;;);"
    #     print("res: ", response)
    #     response = self._parser.decoder.decode(response[9:])
    #     return response.get("payload", {})
    #
    #
    # async def upload_video(self, video_path: str) -> str:
    #     """
    #     Upload a video file to Facebook and return the video ID.
    #
    #     Args:
    #         video_path (str): Path to the video file to upload
    #
    #     Returns:
    #         Optional[int]: Video ID if successful, None if failed
    #     """
    #     if not self._state:
    #         raise LoginError("Client is not logged in yet.")
    #
    #     try:
    #         # Validate file
    #         file_path = Path(video_path)
    #         if not file_path.exists():
    #             raise FBChatError(f"Video file not found: {video_path}")
    #         print("giving path")
    #         file_size = os.path.getsize(file_path)
    #         print("done getting path file size")
    #
    #         # Step 1: Get upload configurations
    #         self.logger.debug("Getting upload configurations...")
    #         video_config = await self._get_video_upload_config()
    #         media_config = await self._get_media_upload_config()
    #
    #         # Step 2: Start upload session
    #         start_result = await self._start_video_upload(file_size, media_config)
    #         video_id = start_result["video_id"]
    #         waterfall_id = start_result["waterfall_id"]
    #
    #         if start_result.get("skip_upload"):
    #             self.logger.info(f"Video already exists, skipping upload. Video ID: {video_id}")
    #             return str(video_id)
    #
    #         self.logger.info(f"Upload session started. Video ID: {video_id}")
    #
    #         # Step 3: Read and upload video file
    #         self.logger.debug("Uploading video data...")
    #
    #         async with aiofiles.open(file_path, 'rb') as f:
    #             video_data = await f.read()
    #             self.logger.info(f"video_data type is: {type(video_data)}")
    #
    #         start_offset = start_result["start_offset"]
    #         end_offset = start_result["end_offset"]
    #
    #         # Resumable upload endpoint
    #         service_name = video_config.get("resumable_service_name", "rupload")
    #         service_domain = video_config.get("resumable_service_domain", "facebook.com")
    #         upload_uuid = generate_uuid().replace("-", "")
    #
    #
    #         # Upload host
    #         upload_url = f"https://{service_name}.{service_domain}/fb_video/{upload_uuid}-{start_offset}-{end_offset}"
    #
    #         await self._prepare_for_upload_chunk(upload_url)
    #
    #         upload_handle = await self._upload_video_chunk(
    #             upload_url,
    #             video_id=video_id,
    #             video_data=(os.path.basename(file_path), video_data, self._state._magic.from_file(file_path)),
    #             waterfall_id=waterfall_id,
    #             start_offset=start_offset,
    #             end_offset=end_offset,
    #             video_size=str(file_size)
    #         )
    #
    #         self.logger.info(f"Video data uploaded. Handle: {upload_handle}...")
    #
    #         # Step 4: Finalize upload
    #         self.logger.info("Finalizing upload...")
    #         finalize_result = await self._finalize_video_upload(
    #             video_id=video_id,
    #             start_offset=start_offset, 
    #             end_offset=end_offset,
    #             upload_handle=upload_handle,
    #             waterfall_id=waterfall_id,
    #             config=media_config,
    #             file_size=file_size
    #         )
    #         self.logger.debug(f"Finalized result: {finalize_result}")
    #
    #         self.logger.info(f"Video upload completed successfully. Video ID: {video_id}")
    #
    #         return str(video_id)
    #
    #     except Exception as e:
    #         raise FBChatError(f"Error uploading video {video_path}: {str(e)}")
    #
    # async def upload_videos( 
    #     self, 
    #     video_paths: List[str], 
    #     max_concurrent: int = 2  # Videos are larger, use fewer concurrent uploads
    # ) -> List[str]:
    #     """
    #     Upload multiple videos concurrently and return list of video IDs.
    #
    #     Args:
    #         video_paths (List[str]): List of video file paths to upload
    #         max_concurrent (int): Maximum number of concurrent uploads (default: 2)
    #
    #     Returns:
    #         List[int]: List of successfully uploaded video IDs
    #     """
    #     if not video_paths:
    #         return []
    #
    #     self.logger.info(f"Starting upload of {len(video_paths)} videos")
    #
    #     # Create semaphore to limit concurrent uploads
    #     semaphore = asyncio.Semaphore(max_concurrent)
    #
    #     async def upload_with_semaphore(path: str, index: int) -> str:
    #         """Upload with semaphore and return (index, video_id)"""
    #         async with semaphore:
    #             self.logger.info(f"Uploading video {index + 1}/{len(video_paths)}: {path}")
    #             video_id = await self.upload_video(path)
    #             return str(video_id)
    #
    #     # tasks for all uploads
    #     tasks = [
    #         upload_with_semaphore(path, idx) 
    #         for idx, path in enumerate(video_paths)
    #     ]
    #
    #     # Wait for all uploads to complete
    #     results = await asyncio.gather(*tasks, return_exceptions=True)
    #
    #     # Process results
    #     video_ids = []
    #     errors = []
    #
    #     for result in results:
    #         if isinstance(result, Exception):
    #             self.logger.error(f"Video upload task failed: {result}")
    #             errors.append(result)
    #         elif result is not None:
    #             video_ids.append(result)
    #         else:
    #             errors.append(f"Upload failed for index {result}")
    #
    #     # Sort by original index
    #
    #     self.logger.info(f"Video upload complete: {len(video_ids)}/{len(video_paths)} successful")
    #
    #     return video_ids







from msgspec import Struct, Raw

class LSResp(Struct, frozen=True, eq=False):
    request_id: int
    """Id of the sent request. Used to keep track of requests"""
    payload: str
    """Kind of like a pointer pointing to raw bytes of the payload in the memory for later decoding"""







"""Commonly received json object in message data"""

from msgspec import Struct, field
from .deltas.custom_type import Value

class MessageData(Struct, frozen=True, omit_defaults=True):
    """A message metadata information, including message senderId (The User that triggered the eevnt), thread folder, message id, timestamp etc."""

    id: str = field(name="messageId")
    """Id of the message"""
    sender_id: str | int = field(name="actorFbId")
    """The user Who sent the message or triggered it"""
    folder: Value = field(name="folderId")
    """The Thread folder location such as Inbox, Archive etc. Usually It is `Inbox`"""
    timestamp: str | int 
    """The timestamp of the message."""
    thread_id: Value = field(name="threadKey")
    """The thread Id where the message was sent"""
    adminText: str = ""
    """Only received when in a group any admin related actions are done such as Kicking , adding memebers, promoting to admin, approving Users to group etc. """
    unsendType: str = "unknown"
    """Wether the message can be unsent or not. Client can only unsent if the value is 'can_unsend'"""





"""
Received Payloads from "ClientPayload" json structure.
Struct classes are used to parse them into Class objects instead of parsing into dict using standard json module.
"""

from typing import Optional, List
from msgspec import Struct, field


from ..messagesData import MessageData
from .custom_type import MentionType, Value
from .attachments_deltas import PayloadAttachments 
from ..message import MessageReaction, MessageRemove, MessageUnsend
from ..notifications import PageNotification
from ..thread_actions import ChangeViwerStatus, ApprovedUser, MuteThread

class MagicWord(Struct, frozen=True, eq=False):
    magic_word: str = field(name="magicWord")
    timestamp: str = field(name="timestampMS")
    emoji: str

class PayloadReplyMessage(Struct, frozen=True, eq=False):
    """"""
    messageMetadata: MessageData 
    body: Optional[str] = None
    messageReply: Optional[Value] = None
    attachments: Optional[List[PayloadAttachments]] = None 
    mentions: MentionType  = field(name="data", default=MentionType())
    participants: tuple = tuple()


class PayloadDeltaReply(Struct, frozen=True, eq=False):
    """"""
    repliedToMessage: Optional[PayloadReplyMessage] = None # main message
    message: Optional[PayloadReplyMessage] = None   # the message `main` message replied to


class PayloadThreadTheme(Struct, frozen=True, eq=False):
    """"""
    # thread_id: Value = field(name="threadKey")
    # theme_id: int = field(name="themeId")
    # fallbackColor: str 
    # gradientColors: List[str]
    # background_image_id: Value = field(name="backgroundAsset")
    # backgroundGradientColors: List[str]
    # titleBarBackgroundColor: str
    # composerBackgroundColor: str
    # titleBarTextColor: str
    # titleBarAttributionColor: str
    # composerInputBackgroundColor: str 
    # composerInputPlaceholderColor: str
    # inboundMessageGradientColors: List[str]
    # titleBarButtonTintColor: str
    # composerTintColor: str
    # composerUnselectedTintColor: str
    # messageTextColor: str
    # inboundMessageTextColor: str
    # deliveryReceiptColor: str
    # tertiaryTextColor: str
    # primaryButtonBackgroundColor: str
    # voiceRecordSoundwaveColor: str
    # accessibility_label: str = field(name="accessibilityLabel")

class PayloadThreadEmoji(Struct, frozen=True, eq=False):
    """"""
    # thread_id: Value = field(name="threadKey")
    # emoji: str


class PayloadMagicWords(Struct, frozen=True, eq=False):
    """"""
    # thread_id: Value = field(name="threadKey")
    # newMagicWords: List[MagicWord]
    # removedMagicWords: List[str]


class PayloadAdminsAdded(Struct, frozen=True, eq=False):
    """"""
    # thread_id: Value = field(name="threadKey")
    # added_admins: List[Value] = field(name="promotedAdmins")

    
class PayloadMoveToArchive(Struct, frozen=True, eq=False):
    """"""

# class MessageUnsend moved to message.py for documentaion

# class MessageReaction moved to message.py 

# class MessageRemove moved to message.py

# class PageNotification moved to fbnotification.py

# class ChangeViwerStatus moved to thread_actions.py

# class ApprovedUser moved to thread_actions.py


class DeltaMessageReply(Struct, frozen=True):
    messageReply: Optional[PayloadDeltaReply] = field(name="deltaMessageReply", default=None)
    replyType: Optional[int] = None
    messageReaction: Optional[MessageReaction] = field(name="deltaMessageReaction", default=None)
    messageUnsend: Optional[MessageUnsend] = field(name="deltaRecallMessageData", default=None)
    messageRemove: Optional[MessageRemove] =  field(name="deltaRemoveMessage", default=None)

    UpdateThreadTheme: Optional[PayloadThreadTheme] = field(name="deltaUpdateThreadTheme", default=None)
    UpdateThreadEmoji: Optional[PayloadThreadEmoji] = field(name="deltaUpdateThreadEmoji", default=None)
    UpdateMagicWords: Optional[PayloadMagicWords] = field(name="deltaUpdateMagicWords", default=None)
    PromoteGroupThreadAdmin: Optional[PayloadAdminsAdded] = field(name="deltaPromoteGroupThreadAdmin", default=None)
    AcceptToGroupThread: Optional[ApprovedUser] = field(name="deltaAcceptGroupThread", default=None)
    muteThread: Optional[MuteThread] = field(name="deltaMuteCallsFromThread", default=None)
    moveThreadToArchive: Optional[PayloadMoveToArchive] = field(name="deltaUpdatePinnedThread", default=None)

    changeViewerStatus: Optional[ChangeViwerStatus] = field(name="deltaChangeViewerStatus", default=None)
    pageNotification: Optional[PageNotification] = field(name="deltaBiiMPageMessageNotification", default=None)




from msgspec import Struct, field
from typing import List, Optional
from .deltas.custom_type import Value


class Asset(Struct, frozen=True):
    """Represents a theme asset (icon or background)"""
    id: str
    """image Id"""
    image: Value
    """Image url"""


class AlternativeTheme(Struct, frozen=True):
    """Alternative theme variant (typically dark mode)"""
    id: str
    accessibility_label: str
    description: Optional[str]
    app_color_mode: str
    composer_background_color: Optional[str]
    background_gradient_colors: List[str]
    is_deprecated: bool
    title_bar_button_tint_color: Optional[str]
    inbound_message_gradient_colors: List[str]
    title_bar_text_color: Optional[str]
    composer_tint_color: Optional[str]
    title_bar_attribution_color: Optional[str]
    composer_input_background_color: Optional[str]
    hot_like_color: Optional[str]
    background_asset: Optional[Asset]
    message_text_color: Optional[str]
    message_border_color: Optional[str]
    message_border_width: Optional[int]
    inbound_message_text_color: Optional[str]
    inbound_message_border_color: Optional[str]
    inbound_message_border_width: Optional[int]
    normal_theme_id: str
    primary_button_background_color: Optional[str]
    title_bar_background_color: Optional[str]
    tertiary_text_color: Optional[str]
    reverse_gradients_for_radial: bool
    reaction_pill_background_color: Optional[str]
    secondary_text_color: Optional[str]
    fallback_color: Optional[str]
    gradient_colors: List[str]
    icon_asset: Optional[Asset]


class Theme(Struct, frozen=True):
    """Main theme information"""
    id: str
    accessibility_label: str
    description: Optional[str]
    app_color_mode: str
    composer_background_color: Optional[str]
    background_gradient_colors: List[str]
    is_deprecated: bool
    title_bar_button_tint_color: Optional[str]
    inbound_message_gradient_colors: List[str]
    title_bar_text_color: Optional[str]
    composer_tint_color: Optional[str]
    title_bar_attribution_color: Optional[str]
    composer_input_background_color: Optional[str]
    hot_like_color: Optional[str]
    background_asset: Optional[Asset]
    message_text_color: Optional[str]
    message_border_color: Optional[str]
    message_border_width: Optional[int]
    inbound_message_text_color: Optional[str]
    inbound_message_border_color: Optional[str]
    inbound_message_border_width: Optional[int]
    primary_button_background_color: Optional[str]
    title_bar_background_color: Optional[str]
    tertiary_text_color: Optional[str]
    reverse_gradients_for_radial: bool
    reaction_pill_background_color: Optional[str]
    secondary_text_color: Optional[str]
    fallback_color: Optional[str]
    gradient_colors: List[str]
    normal_theme_id: str
    icon_asset: Optional[Asset]
    alternative_themes: List[AlternativeTheme]


class ThreadThemes(Struct, frozen=True):
    """Container for messenger thread themes"""
    themes: List[Theme] = field(name="messenger_thread_themes")


class ThemeData(Struct, frozen=True):
    """Root theme data structure"""
    data: ThreadThemes





"""
GraphQL utilities for fbchat-muqit.

This module provides utilities for handling GraphQL queries, responses,
and JSON parsing for Facebook's GraphQL API.
"""

import json
from typing import List, Dict, Any, Optional
from enum import IntEnum

import msgspec

# Use the existing logging and exception systems
from .logging.logger import get_logger
from .exception.errors import (
    FBChatError, 
    FacebookAPIError, 
    ValidationError, 
    handle_exceptions
)


class FacebookErrorCode(IntEnum):
    """Facebook API error codes."""
    NOT_LOGGED_IN = 1357001
    REFRESH_COOKIES = 1357004
    INVALID_PARAMS_1 = 1357031
    INVALID_PARAMS_2 = 1545010
    INVALID_PARAMS_3 = 1545003


class GraphQLError(msgspec.Struct, frozen=True, eq=False):
    """Represents a GraphQL error."""
    code: Optional[int] = None
    message: Optional[str] = None
    severity: Optional[str] = None


class QueryRequest(msgspec.Struct, frozen=True, eq=False):
    """Represents a GraphQL query request."""
    priority: int = 0
    query: Optional[str] = None
    query_id: Optional[str] = None
    doc: Optional[str] = None
    doc_id: Optional[str] = None
    query_params: Dict[str, Any] | None = dict()


class GraphQLProcessor:
    """Handles GraphQL query processing and response parsing."""
    
    def __init__(self):
        self.logger = get_logger()
        self.decoder = msgspec.json.Decoder()
    
    @handle_exceptions(ValidationError)
    def parse_json_stream(self, content: str) -> List[Dict[str, Any]]:
        """
        Parse multiple concatenated JSON objects in a single string.
        
        Args:
            content: String containing one or more JSON objects
            
        Returns:
            List of parsed JSON objects
            
        Raises:
            ValidationError: If JSON parsing fails
        """
        if not content.strip():
            self.logger.trace("Empty content provided to parse_json_stream")
            return []
            
        results: List[Dict[str, Any]] = []
        content = content.strip()
        idx = 0

        self.logger.trace(f"Parsing JSON stream with {len(content)} characters")

        decoder = json.JSONDecoder()

        while idx < len(content):
            # Skip whitespace
            while idx < len(content) and content[idx].isspace():
                idx += 1
                
            if idx >= len(content):
                break
                
            try:
                obj, offset = decoder.raw_decode(content, idx)
                results.append(obj)
                idx += offset
                self.logger.trace(f"Successfully parsed JSON object at position {idx - offset}")
            except json.JSONDecodeError as e:
                self.logger.warning(f"JSON decode error at position {idx}: {e}")
                break
            except Exception as e:
                self.logger.error(f"Unexpected error parsing JSON stream: {e}")
                raise ValidationError(
                    f"Failed to parse JSON stream: {e}",
                    details={'content_length': len(content), 'position': idx}
                ) from e

        self.logger.debug(f"Parsed {len(results)} JSON objects from stream")
        return results


    def process_normal_response(self, content: str)->Dict[str, Any]:
        try:
            cleaned_content = self.strip_json_cruft(content)
            return self.decoder.decode(cleaned_content)
        except (ValueError, msgspec.DecodeError) as e:
            raise ValidationError(
                "No valid JSON found in response",
                details={'content_preview': content[:100] + '...' if len(content) > 100 else content}
            ) from e
   

    @handle_exceptions(ValidationError)
    def strip_json_cruft(self, content: str) -> str:
        """
        Remove Facebook's JSON cruft like 'for(;;);' that precedes responses.
        
        Args:
            content: Raw response content
            
        Returns:
            Cleaned JSON string
            
        Raises:
            ValidationError: If no valid JSON is found
        """
        if not content:
            raise ValidationError("Empty content provided to strip_json_cruft")
            
        self.logger.trace(f"Stripping JSON cruft from {len(content)} character response")
        
        # Find the first opening brace
        try:
            start_idx = content.index("{")
            cleaned_content = content[start_idx:]
            cruft_length = start_idx
            if cruft_length > 0:
                cruft = content[:start_idx]
                self.logger.debug(f"Removed {cruft_length} characters of cruft: {cruft!r}")
            return cleaned_content
        except ValueError as e:
            raise ValidationError(
                "No valid JSON found in response",
                details={'content_preview': content[:100] + '...' if len(content) > 100 else content}
            ) from e

    def queries_to_json(self, *queries: QueryRequest) -> str:
        """
        Convert QueryRequest objects to JSON string format.
        
        Args:
            queries: Variable number of QueryRequest objects
            
        Returns:
            JSON string representation of queries
        """
        self.logger.debug(f"Converting {len(queries)} queries to JSON")
        
        rtn = {}
        for i, query in enumerate(queries):
            if query.query:
                rtn[f"q{i}"] = {
                    "priority": query.priority,
                    "q": query.query,
                    "query_params": query.query_params
                }
                self.logger.trace(f"Query {i}: GraphQL query with {len(query.query)} characters")
            elif query.query_id:
                rtn[f"q{i}"] = {
                    "query_id": query.query_id,
                    "query_params": query.query_params
                }
                self.logger.trace(f"Query {i}: Query ID {query.query_id}")
            elif query.doc:
                rtn[f"q{i}"] = {
                    "doc": query.doc,
                    "query_params": query.query_params
                }
                self.logger.trace(f"Query {i}: Document with {len(query.doc)} characters")
            elif query.doc_id:
                rtn[f"q{i}"] = {
                    "doc_id": query.doc_id,
                    "query_params": query.query_params
                }
                self.logger.trace(f"Query {i}: Document ID {query.doc_id}")
            
        json_str = json.dumps(rtn)
        self.logger.debug(f"Generated JSON query string with {len(json_str)} characters")
        return json_str



    @handle_exceptions(FacebookAPIError)
    def handle_payload_error(self, payload: Dict[str, Any]) -> None:
        """
        Handle Facebook API payload errors.
        
        Args:
            payload: The JSON payload to check for errors
            
        Raises:
            FacebookAPIError: For critical errors
        """
        if "error" not in payload:
            return
            
        error_code = payload["error"]
        self.logger.warning(f"Facebook payload error detected: {error_code}")
        
        error_details = {
            'error_code': error_code,
            'payload': payload
        }
        
        if error_code == FacebookErrorCode.NOT_LOGGED_IN:
            self.logger.error("Not logged into Facebook")
            raise FacebookAPIError(
                "Not logged in - please authenticate",
                error_code=error_code,
                details=error_details
            )
        elif error_code == FacebookErrorCode.REFRESH_COOKIES:
            self.logger.error("Cookies need to be refreshed")
            raise FacebookAPIError(
                "Please refresh your authentication cookies",
                error_code=error_code,
                details=error_details
            )
        elif error_code in (
            FacebookErrorCode.INVALID_PARAMS_1,
            FacebookErrorCode.INVALID_PARAMS_2,
            FacebookErrorCode.INVALID_PARAMS_3
        ):
            self.logger.error(f"Invalid parameters provided (error: {error_code})")
            raise FacebookAPIError(
                "Invalid parameters provided",
                error_code=error_code,
                details=error_details
            )
        else:
            self.logger.error(f"Unknown Facebook API error: {error_code}")
            raise FacebookAPIError(
                f"Facebook API error: {error_code}",
                error_code=error_code,
                details=error_details
            )

    @handle_exceptions(FacebookAPIError)
    def handle_graphql_errors(self, response: Dict[str, Any]) -> None:
        """
        Handle GraphQL-specific errors in the response.
        
        Args:
            response: GraphQL response object
            
        Raises:
            FacebookAPIError: For critical GraphQL errors
        """
        errors = []
        
        if response.get("error"):
            errors = [response["error"]]
        elif response.get("errors"):
            errors = response["errors"]
            
        if not errors:
            return
        
        self.logger.warning(f"GraphQL errors detected: {len(errors)} error(s)")
        
        # Process the first error (most critical)
        error_data = errors[0]
        error = GraphQLError(
            code=error_data.get("code"),
            message=error_data.get("message", "Unknown GraphQL error"),
            severity=error_data.get("severity")
        )
        
        self.logger.error(f"GraphQL error - Code: {error.code}, Message: {error.message}")
        
        # Raise exception for critical errors
        if error.severity == "CRITICAL" or error.code:
            raise FacebookAPIError(
                f"GraphQL error: {error.message}",
                error_code=error.code,
                details={
                    'graphql_errors': errors,
                    'severity': error.severity,
                    'response': response
                }
            )

    @handle_exceptions(FBChatError)
    def process_response(self, content: str) -> List[Optional[Dict[str, Any]]]:
        """
        Process Facebook's JSON response and extract structured data.
        
        Args:
            content: Raw JSON string from Facebook's API
            
        Returns:
            List of parsed response objects (None for invalid responses)
            
        Raises:
            FBChatError: If parsing fails or critical errors occur
        """
        if not content:
            self.logger.debug("Empty content provided to process_response")
            return []
        
        # self.logger.info(f"Processing Facebook response ({len(content)} characters)")
        
        # Clean the content and parse JSON objects
        try:
            cleaned_content = self.strip_json_cruft(content)
            parsed_objects = self.parse_json_stream(cleaned_content)
        except Exception as e:
            self.logger.error(f"Failed to parse response: {e}")
            raise FBChatError(f"Error parsing response: {e}") from e
        
        if not parsed_objects:
            self.logger.warning("No valid JSON objects found in response")
            return []
        
        self.logger.debug(f"Processing {len(parsed_objects)} parsed objects")
        
        results: List[Optional[Dict[str, Any]]] = []
        for obj_idx, obj in enumerate(parsed_objects):
            try:
                if "error_results" in obj:
                    continue
                # Handle payload-level errors
                self.handle_payload_error(obj)
                
                for key, value in obj.items():
                    self.logger.trace(f"Processing response key: {key}")
                    
                    # Handle GraphQL errors
                    self.handle_graphql_errors(value)
                    
                    # FIXED: Extract index and store result (matching original logic)
                    try:
                        index = int(key[1:])  # Extract numeric part (e.g., "q0" -> 0)
                            
                         # Extend results list if needed
                        while len(results) <= index:
                            results.append(None)
                            
                        # Store the result
                        if "response" in value:
                            results[index] = value["response"]
                            self.logger.trace(f"Extracted 'response' data for query {index}")
                        elif "data" in value:
                            results[index] = value["data"]
                            self.logger.trace(f"Extracted 'data' data for query {index}")
                        else:
                            results[index] = value
                            self.logger.trace(f"Using raw value for query {index}")
                                
                    except (ValueError, IndexError) as e:
                        self.logger.warning(f"Invalid query key format: {key} - {e}")
                        continue
                    else:
                        # Handle non-query keys if needed
                        self.logger.trace(f"Ignoring non-query key: {key}")
                        
            except FBChatError:
                raise  # Re-raise our custom exceptions (already logged)
            except Exception as e:
                self.logger.error(f"Error processing response object {obj_idx}: {e}")
                continue
        
        processed_count = sum(1 for r in results if r is not None)
        self.logger.info(f"Successfully processed {processed_count}/{len(results)} responses")
        
        return results


# Global processor instance for backward compatibility
_global_processor: Optional[GraphQLProcessor] = None


def get_processor() -> GraphQLProcessor:
    """Get the global GraphQL processor instance."""
    global _global_processor
    if _global_processor is None:
        _global_processor = GraphQLProcessor()
    return _global_processor


# Factory functions for creating QueryRequest objects
def from_query(query: str, params: Optional[Dict[str, Any]] = None) -> QueryRequest:
    """Create a QueryRequest from a GraphQL query string."""
    logger = get_logger()
    logger.trace(f"Creating QueryRequest from query ({len(query)} chars)")
    return QueryRequest(priority=0, query=query, query_params=params or {})


def from_query_id(query_id: str, params: Optional[Dict[str, Any]] = None) -> QueryRequest:
    """Create a QueryRequest from a query ID."""
    logger = get_logger()
    logger.trace(f"Creating QueryRequest from query_id: {query_id}")
    return QueryRequest(query_id=query_id, query_params=params or {})


def from_doc(doc: str, params: Optional[Dict[str, Any]] = None) -> QueryRequest:
    """Create a QueryRequest from a document string."""
    logger = get_logger()
    logger.trace(f"Creating QueryRequest from document ({len(doc)} chars)")
    return QueryRequest(doc=doc, query_params=params or {})


def from_doc_id(doc_id: str, params: Optional[Dict[str, Any]] = None) -> QueryRequest:
    """Create a QueryRequest from a document ID."""
    logger = get_logger()
    logger.trace(f"Creating QueryRequest from doc_id: {doc_id}")
    return QueryRequest(doc_id=doc_id, query_params=params or {})


# Backward compatibility functions (using global processor)
def parse_json_stream(content: str) -> List[Dict[str, Any]]:
    """Parse multiple concatenated JSON objects (backward compatibility)."""
    return get_processor().parse_json_stream(content)


def strip_json_cruft(content: str) -> str:
    """Remove JSON cruft from content (backward compatibility)."""
    return get_processor().strip_json_cruft(content)


def queries_to_json(*queries: QueryRequest) -> str:
    """Convert queries to JSON string (backward compatibility)."""
    return get_processor().queries_to_json(*queries)


def handle_payload_error(payload: Dict[str, Any]) -> None:
    """Handle payload errors (backward compatibility)."""
    return get_processor().handle_payload_error(payload)


def handle_graphql_error(response: Dict[str, Any]) -> None:
    """Handle GraphQL errors (backward compatibility - note: singular 'error' for compatibility)."""
    return get_processor().handle_graphql_errors(response)


def response_to_json(content: str) -> List[Optional[Dict[str, Any]]]:
    """Process response content (backward compatibility)."""
    return get_processor().process_response(content)



def extract_message_id_raw(payload: str) -> str | None:
    marker = "replaceOptimsiticMessage"
    idx = payload.find(marker)
    if idx == -1:
        return None
    # Find next "mid." after the marker
    mid_start = payload.find("mid.", idx)
    if mid_start == -1:
        return None
    mid_end = payload.find('"', mid_start)
    return payload[mid_start:mid_end] if mid_end != -1 else None




"""
fbchat_muqit/logging/logger.py - Enhanced Logging System
"""

import json
import logging
import sys
from datetime import datetime
from enum import Enum
from typing import Any, Optional, Union


class LogLevel(Enum):
    TRACE = 5
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    API_REQUEST = 15
    API_RESPONSE = 16
    MESSAGE = 25
    CONNECTION = 35


class ColoredFormatter(logging.Formatter):
    COLORS = {
        'TRACE': '\033[90m',
        'DEBUG': '\033[94m',
        'API_REQUEST': '\033[96m',
        'API_RESPONSE': '\033[36m',
        'INFO': '\033[92m',
        'MESSAGE': '\033[93m',
        'WARNING': '\033[93m',
        'CONNECTION': '\033[95m',
        'ERROR': '\033[91m',
        'CRITICAL': '\033[91m\033[1m',
        'RESET': '\033[0m'
    }

    EMOJIS = {
        'TRACE': '🔍',
        'DEBUG': '🐛',
        'API_REQUEST': '📤',
        'API_RESPONSE': '📥',
        'INFO': '📜',
        'MESSAGE': '💬',
        'WARNING': '⚠️',
        'CONNECTION': '🔗',
        'ERROR': '❌',
        'CRITICAL': '💥'
    }

    def format(self, record):
        color = self.COLORS.get(record.levelname, '')
        emoji = self.EMOJIS.get(record.levelname, '📝')
        reset = self.COLORS['RESET']
        timestamp = datetime.fromtimestamp(record.created).strftime('%H:%M:%S.%f')[:-3]
        message = super().format(record)
        caller = f" [{record.filename}:{record.lineno}]" if record.levelno <= logging.DEBUG else ""
        return f"{color}{emoji} {timestamp} | {record.levelname:6} | {record.name}{caller} | {message}{reset}"


class JSONFormatter(logging.Formatter):
    def format(self, record):
        entry = {
            'timestamp': datetime.fromtimestamp(record.created).isoformat(),
            'level': record.levelname,
            'logger': record.name,
            'message': record.getMessage(),
            'module': record.module,
            'function': record.funcName,
            'line': record.lineno,
        }
        if record.exc_info:
            entry['exception'] = self.formatException(record.exc_info)
        if hasattr(record, 'extra_data'):
            entry['extra'] = record.extra_data #type: ignore
        return json.dumps(entry, ensure_ascii=False)


class FBChatLogger:
    def __init__(
        self,
        name: str = "fbchat-muqit",
        level: Union[int, str, LogLevel] = LogLevel.INFO,
        console_output: bool = True,
        enable_colors: bool = True,
        log_api_requests: bool = True,
        log_api_responses: bool = False,
        sensitive_fields: Optional[list] = None
    ):
        self.name = name
        self.sensitive_fields = sensitive_fields or ['password', 'token', 'access_token', 'session', "email", "fb_dtsg", "lsd"]
        self.log_api_requests = log_api_requests
        self.log_api_responses = log_api_responses

        self.logger = logging.getLogger(name)
        self.logger.setLevel(self._convert_level(level))
        self._add_custom_levels()
        self.logger.handlers.clear()

        if console_output:
            self._setup_console_handler(enable_colors)

    def _convert_level(self, level):
        if isinstance(level, LogLevel):
            return level.value
        if isinstance(level, str):
            return getattr(logging, level.upper(), logging.INFO)
        return level

    def _add_custom_levels(self):
        for lvl in LogLevel:
            logging.addLevelName(lvl.value, lvl.name)
            method = lvl.name.lower()
            if not hasattr(self.logger, method):
                setattr(self.logger, method, lambda msg, *a, lvl=lvl.value, **kw: self.logger.log(lvl, msg, *a, **kw))

    def _setup_console_handler(self, enable_colors: bool):
        handler = logging.StreamHandler(sys.stdout)
        formatter = ColoredFormatter() if enable_colors and sys.stdout.isatty() else logging.Formatter('%(asctime)s | %(levelname)-12s | %(name)s | %(message)s', '%H:%M:%S')
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

    def _sanitize_data(self, data: Any) -> Any:
        if isinstance(data, dict):
            return {k: ("***REDACTED***" if any(s in k.lower() for s in self.sensitive_fields) else self._sanitize_data(v)) for k, v in data.items()}
        if isinstance(data, (list, tuple)):
            return [self._sanitize_data(i) for i in data]
        return data

    # ─────────── Logging Methods ───────────

    def debug(self, msg: str, *a, **kw): self.logger.debug(msg, stacklevel=2, *a, **kw)
    def info(self, msg: str, *a, **kw): self.logger.info(msg, stacklevel=2, *a, **kw)
    def warning(self, msg: str, *a, **kw): self.logger.warning(msg, stacklevel=2, *a, **kw)
    def trace(self, msg: str, *a, **kw): self.logger.log(LogLevel.TRACE.value, msg, stacklevel=2, *a, **kw)

    def error(self, msg: str, *a, exc_info=None, **kw):
        prefix = "❌ " if not msg.startswith(("❌", "⚠️", "💥")) else ""
        self.logger.error(prefix + msg, *a, exc_info=exc_info, stacklevel=2, **kw)

    def critical(self, msg: str, *a, **kw): self.logger.critical("💥 " + msg, *a, **kw)

    def exception(self, exc: Exception, context: str = ""):
        emoji = getattr(exc, "emoji", "❌")
        msg = f"{emoji} {exc.__class__.__name__}: {exc}"
        ctx = f" ({context})" if context else ""
        self.logger.error(f"{msg}{ctx}", exc_info=True, extra={'extra_data': {'context': context}})

    # ─────────── Specialized Logging ───────────

    def log_api_request(self, method: str, url: str, data: Any = None, headers: Any = None):
        if not self.log_api_requests: return
        sanitized = self._sanitize_data(data) if data else None
        self.logger.log(LogLevel.API_REQUEST.value, f"📤 API Request: {method} {url}", extra={'extra_data': {'data': sanitized}})

    def log_api_response(self, status: int, url: str, resp: Any = None, duration: float | None = None):
        if not self.log_api_responses: return
        sanitized = self._sanitize_data(resp) if resp else None
        dur = f" ({duration:.3f}s)" if duration else ""
        self.logger.log(LogLevel.API_RESPONSE.value, f"📥 API Response: {status} {url}{dur}", extra={'extra_data': {'response': sanitized, 'duration': duration}})

    def log_connection_event(self, event: str, details: Any = None):
        self.logger.log(LogLevel.CONNECTION.value, f"🔗 Connection: {event}", extra={'extra_data': {'details': details}})

    def log_message_event(self, event_type: str, data: Any = None):
        sanitized = self._sanitize_data(data) if data else None
        self.logger.log(LogLevel.MESSAGE.value, f"💬 Message Event: {event_type}", extra={'extra_data': {'data': sanitized}})


# ─────────── Global Utilities ───────────

global_logger: Optional[FBChatLogger] = None

def get_logger() -> FBChatLogger:
    global global_logger
    if global_logger is None:
        global_logger = FBChatLogger()
    return global_logger

def setup_logger(level: Union[int, str, LogLevel] = LogLevel.INFO, console_output: bool = True, enable_colors: bool = True, **kw) -> FBChatLogger:
    global global_logger
    global_logger = FBChatLogger(level=level, console_output=console_output, enable_colors=enable_colors, **kw)
    return global_logger

def set_log_level(level: Union[int, str, LogLevel]):
    log = get_logger()
    val = level.value if isinstance(level, LogLevel) else getattr(logging, str(level).upper(), logging.INFO)
    log.logger.setLevel(val)

def enable_debug(): set_log_level(LogLevel.DEBUG)
def enable_trace(): set_log_level(LogLevel.TRACE)
def disable_logging(): set_log_level(logging.CRITICAL + 10)

def get_current_log_level() -> str:
    lvl = get_logger().logger.getEffectiveLevel()
    for l in LogLevel:
        if l.value == lvl:
            return l.name
    return logging.getLevelName(lvl)




from .state import State
from .client import Client
from .facebook.client import  FacebookClient
from .messenger.client import MessengerClient
from .events.dispatcher import EventType, EventDispatcher, EventCallback
from .models import * 


__title__ = "fbchat-muqit"
__version__ = "1.1.5"
__description__ = "A powerful Facebook Messenger Chat API to interact with Facebook and Messenger. Easy to use fast and efficient library"

__copyright__ = "Copyright 2024 by Muhammad MuQiT"

__license__ = "GPL-V3.0"

__author__ = "Muhammad MuQiT"
__email__ = "togashiyuuta1111@gmail.com"

__all__ = [
    # Core classes
    "Client",
    "State",
    "FacebookClient",
    "MessengerClient",
    
    # Thread related
    "ThreadType",
    "ThreadFolder",
    "Thread",
    
    # User
    "User",
    
    # Attachments
    "Attachment",
    "AttachmentType",
    "ImageAttachment",
    "VideoAttachment",
    "GifAttachment",
    "StickerAttachment",
    "AudioAttachment",
    "FileAttachment",
    "LocationAttachment",
    "PostAttachment",
    "SharedAttachment",
    "ReelAttachment",
    "ProfileAttachment",
    "ProductAttachment",
    "ExternalAttachment",
    "Post",
    "Author",
    "GroupInfo",
    "Media",
    "Image",
    "Dimension",
    
    # Thread Actions
    "ApprovalMode",
    "ApprovalQueue",
    "AdminAdded",
    "AdminRemoved",
    "ApprovedUser",
    "addedParticipant",
    "ChangeViwerStatus",
    "JoinableMode",
    "ParticipantsAdded",
    "ParticipantLeft",
    "ThreadAction",
    "ThreadDelete",
    "ThreadEmoji",
    "ThreadFolderMove",
    "ThreadMagicWord",
    "ThreadMessagePin",
    "ThreadMessageUnPin",
    "ThreadMessageSharing",
    "ThreadMuteSettings",
    "ThreadName",
    "ThreadNickname",
    "ThreadTheme",
    "MuteThread",
    "ForcedFetch",
    
    # Typing
    "Typing",
    
    # Timestamps
    "ReadReceipt",
    "DeliveryReceipt",
    "MarkFolderSeen",
    "MarkRead",
    "MarkUnread",
    
    # Messages
    "Message",
    "Mention",
    "Mentions",
    "MessageType",
    "MessageRemove",
    "MessageReaction",
    "MessageUnsend",
    "Reaction",
    "EmojiSize",
    
    # Message Data
    "MessageData",
    
    # Notifications
    "PokeNotification",
    "PageNotification",
    "FriendRequestState",
    "friendUpdated",
    "friendRequestList",
    
    # Presence
    "Presence",
    "UserStatus",
    
    # Themes
    "Theme",
    "AlternativeTheme",
    "Asset",

    # Event 
    "EventType",
    "EventDispatcher",
    "EventCallback",

    # Message Response 
    "MessageSearchStatus",
    "MessageSearchResult",
]




# fbchat_muqit/client.py 
# main client
from __future__ import annotations

from msgspec.json import Decoder

from fbchat_muqit.models.mqtt_response.response import LSResp

__all__ = ["Client"]


import asyncio
import sys
from typing import Optional

# Base Classes
from .facebook.client import FacebookClient
from .messenger.client import MessengerClient
from .state import State
from .muqit import Mqtt
from .realtime import FacebookRealtime
from .events.dispatcher import EventDispatcher, EventType

from .logging.logger import FBChatLogger, setup_logger, disable_logging
from .exception.errors import FBChatError
from .models.deltas.parser import MessageParser, ParsedEvent

class Client(EventDispatcher, FacebookClient, MessengerClient):
        
    def __init__(
            self,
            cookies_file_path: str,
            userAgent: Optional[str] = None,
            proxy: Optional[str] = None,
            log_level = "INFO",
            disable_logs = False,
            online = True
            ):
        
        self._setup_windows_compatibility()
        super().__init__()

        self._state: Optional[State] = None
        self._uid: str = ""
        self._name: str = ""
        self._mqtt: Optional[Mqtt] = None
        self._parser = MessageParser(self.logger)
        self._realtime: Optional[FacebookRealtime] = None
        self._events_queue: asyncio.Queue[ParsedEvent] = asyncio.Queue(maxsize=1000)

        
        self._cookies_file_path = cookies_file_path
        self._userAgent = userAgent
        self._proxy = proxy
        self._online = online
        self.logger: FBChatLogger = setup_logger(log_level)

        self._listening: bool = False 
        
        if disable_logs:
            disable_logging()

    @property
    def uid(self):
        """The Facebook user ID of the client."""
        return self._uid
    @property
    def name(self):
        """The Facebook name of the client."""
        return self._name

    async def __aenter__(self):
     # Reopen session if it was closed
        if not self._state:
            state: State = await State.from_json_cookies(self._cookies_file_path, self._userAgent ,self._proxy)
            self._state = state 
        # Refresh every 1 hour
        # self._state.enable_auto_refresh(interval=3600)
        # self.logger.debug("Started auto state refresh")
        self._uid = self._state.user_id
        self._name = self._state.user_name

        self._parser = MessageParser(self.logger)
        self._parse_ls_resp = Decoder(type=LSResp, strict=False)

        # Ensure logged in
        if not self._state._is_logged:
            self.logger.info("User is not logged in!")

        if self._state.user_name:
            self.logger.info(f"Logged in as {self.name} ({self.uid})")
        return self


    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._state:
            await self._state.close()
            self._state = None 

        await self.stop_listening()

    @staticmethod
    def _setup_windows_compatibility():
        # Setup Windows asyncio compatibility
        if sys.platform == 'win32':
            try:
                # Check current policy
                current_policy = asyncio.get_event_loop_policy()
                # Switch to SelectorEventLoop if using ProactorEventLoop
                if hasattr(asyncio, 'WindowsProactorEventLoopPolicy') and \
                   isinstance(current_policy, asyncio.WindowsProactorEventLoopPolicy):
                    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                # Ensure event loop exists
                try:
                    asyncio.get_event_loop()
                except RuntimeError:
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
            except Exception:
                pass

    async def start_listening(self):
        """Start listening from an external event loop.
        
        Raises:
            FBchatException: If request failed
        """
        if not self._state:
            raise FBChatError("Failed to get session")
        
        self._mqtt = await Mqtt.connect(
                state=self._state,
                chat_on=self._online,
                foreground=self._online,
                message_handler=self._handle_mqtt_messages
            )

        self._realtime = await FacebookRealtime.connect(self._state, self._handle_realtime_messages)

        if self._online:
            await self._mqtt.set_chat_on(self._online)
            await self._mqtt.set_foreground(self._online)
        self._listening = True
        # starting dispatcher
        asyncio.create_task(self._dispatch_mqtt_message())


    async def stop_listening(self):
        """Stop the listening mqtt and realtime loop."""
        if self._mqtt:
            await self._mqtt.stop()
        if self._realtime:
            await self._realtime.stop()

        self._listening = False
        self._mqtt = None
        self._realtime = None


    async def listen(self): 
        """Starts listening to events Blockingly"""
        await self.start_listening()

        await self.dispatch(EventType.LISTENING)

        try:
            while self._listening:
                await asyncio.sleep(3600)
        except asyncio.CancelledError:
            self.logger.debug("Client stopped listening!")
            raise 

        except Exception as e:
            self.logger.error(f"Error in listen loop: {e}")
            raise FBChatError("Listening failed", original_exception=e)

    
    async def _dispatch_mqtt_message(self):
        """Dispatches Parsed Event data from Queue"""
        while self._listening:
            parsedEvent = await self._events_queue.get()
            try:
                await self.dispatch(parsedEvent.eventType, *parsedEvent.args)
            except Exception as e:
                self.logger.error(f"Failed to dispatch Event ", exc_info=e)

            finally:
                self._events_queue.task_done()

    
    async def _handle_mqtt_messages(self, topic: str, payload: bytes):
        """Handles and Parses incoming payloads and putting them in Queue"""
        try:
            if topic == "/ls_resp":
                # only received payloads if any payloads were published to /ls_req
                data = self._parse_ls_resp.decode(payload)
                fut = self._pending_requests.pop(data.request_id, None)
                if fut and not fut.done():
                    fut.set_result(data)
                return

            elif topic == "/t_ms" and b'deltas' in payload:
                try:
                    eventData = self._parser.parse_t_ms(payload)
                    for e in eventData:
                        if e:
                            await self._events_queue.put(e)
                except Exception as e:
                    self.logger.error(f"Failed to parse /t_ms deltas: {e}")

            else:
                eventdata = self._parser.parse_all(topic, payload)
                if eventdata:
                    await self._events_queue.put(eventdata)
        except Exception as e:
                self.logger.error(f"Failed to parse payloads ftom Topic: {topic} payload: {payload}", exc_info=e)
    


    async def _handle_realtime_messages(self, topic, payload):
        pass 

    async def start(self):
        """Initiates `Client` class and logins to account. But doesn't listens to event. To listen to events call `listen()` method."""
        await self.__aenter__()

    async def close(self):
        """Closes and clears all connections of Client"""
        await self.__aexit__(None, None, None)


    def run(self):
        """Blocking Call listens to incoming events"""
        try:
            asyncio.run(self._runner())
        except KeyboardInterrupt:
            self.logger.info("Interrupted By the User!")
        except Exception as e:
            self.logger.exception(e)

    async def _runner(self):
        await self.start()
        try:
            await self.listen()
        finally:
            await self.close()



    async def download(self, url: str, filename: str):
        """
        Download a image, video sent from messenger.

        Args:
            url: Url of the image, video.
            filename: name of the file.
        """
        if self._state:
            await self._state.download_file(url, filename)



import random, re
import json

from aiohttp import ClientSession, CookieJar
from aiohttp_socks import ProxyConnector
from typing import Tuple, Optional
from yarl import URL
from http.cookies import SimpleCookie

from ..logging.logger import get_logger 
from ..exception.errors import FBChatError, ValidationError
logger = get_logger()

def get_session(cookie_jar: Optional[CookieJar] = None, proxy: Optional[str] = None)-> ClientSession:
    if proxy:
        connector = ProxyConnector.from_url(proxy)
    else:
        connector = None

    return ClientSession(
            cookie_jar=cookie_jar,
            connector=connector
            )



def load_json_cookies(json_path: str) -> CookieJar:
    """Load and format cookies from fbstate.json for aiohttp or requests sessions."""
    with open(json_path, "r", encoding="utf-8") as f:
        data = json.load(f)

    if not data or not isinstance(data, list):
        raise FBChatError("Invalid fbstate format. Expected a list of cookie objects.")

    key_name = "key" if "key" in data[0] else "name"

    jar = CookieJar()
    for cookie in data:
        name = cookie.get(key_name)
        value = cookie.get("value")
        path = cookie.get("path", "/")
        if not name or value is None:
            continue

        # Prepare separate SimpleCookie for each domain
        for domain_url in ("https://www.facebook.com", "https://www.messenger.com", "https://rupload-ccu1-2.up.facebook.com"):
            domain = ".messenger.com" if "messenger" in domain_url else ".facebook.com"
            if "rupload" in domain_url:
                domain = "up.facebook.com"
            sc = SimpleCookie()
            sc[name] = value
            sc[name]["domain"] = domain  # Facebook accepts dot prefix
            sc[name]["path"] = path
            if "expires" in cookie:
                sc[name]["expires"] = cookie["expires"]
            # stores cookiex for each url
            jar.update_cookies(sc, URL(domain_url))

    logger.info(f" ✅ Succssfully loaded Cookies into Jar from {json_path}")
    return jar


def get_user_id(session: ClientSession)-> str:
        cookies = session.cookie_jar.filter_cookies(URL("https://www.facebook.com"))
        CookiesDict = {name: cookie.value for name, cookie in cookies.items()}
        return str(CookiesDict.get("c_user"))


def client_id_factory()-> str:
      return hex(int(random.random() * 2 ** 31))[2:]

def save_html(html):
    with open("./test.html", "w") as f:
        f.write(html)



def extract_tokens_from_html(html: str)->Tuple:
    """Extracts fb_dtsg, client_revision, mqttAppID etc. from HTML response to Login in facebook"""

    fb_dtsg = re.search(r'"DTSGInitialData".*?"token":"(.*?)"', html)
    if fb_dtsg:
        fb_dtsg = fb_dtsg.group(1)
    else:
        raise ValidationError("'fb_dtsg' token not found.")

    pattern = r'"DTSGInitData"(?:\s*,\s*\[\])?(?:\s*,\s*)\{[^}]*"async_get_token"\s*:\s*"([^"]+)"[^}]*\}'
        
    fb_dtsg_ag = re.search(pattern, html)
    if fb_dtsg_ag:
        fb_dtsg_ag = fb_dtsg_ag.group(1)
    else:
        raise ValidationError("'async_get_token' not found.")

    lsd_token = re.search(r'"LSD"\s*,\s*\[\s*\]\s*,\s*\{\s*"token"\s*:\s*"([A-Za-z0-9_-]+)"', html)
    if lsd_token:
        lsd_token = lsd_token.group(1)

    jazoest = "2" + str(sum(ord(c) for c in fb_dtsg))
    jazoest_async =  "2" + str(sum(ord(c) for c in fb_dtsg_ag))


    clientRevision = re.search(r'client_revision":(\d+)', html)
    if clientRevision:
        clientRevision = int(clientRevision.group(1))

    clientID = re.search(r'\["MqttWebDeviceID".*?"clientID"\s*:\s*"([a-f0-9\-]+)"', html)
    if clientID:
        clientID = clientID.group(1)

    mqttAppID = re.search(r'\["MqttWebConfig".*?"appID"\s*:\s*(\d+)', html)
    if mqttAppID:
        mqttAppID = mqttAppID.group(1)

    userAppID = re.search(r'\["CurrentUserInitialData".*?"APP_ID"\s*:\s*"(\d+)"', html)
    if userAppID:
        userAppID = userAppID.group(1)

    # Extracting Mqtt endpoint for facebook 
    mqttEndpoint = re.search(r'"endpoint"\s*:\s*"([^"]*?region=([a-zA-Z0-9_-]+)[^"]*)"', html)
    if not mqttEndpoint:
        raise ValueError("Mqtt Endpoint not found!")
    endpoint = mqttEndpoint.group(1).encode().decode('unicode_escape')
    region = mqttEndpoint.group(2)

    user_name = re.search(r'"NAME"\s*:\s*"([^"]+)"', html)
    if user_name:
        user_name = user_name.group(1)
        logger.debug(f"User name: {user_name}")

    logger.debug(f"fb dtag: {fb_dtsg}") # used for post requests
    logger.debug(f"fb dtsg async: {fb_dtsg_ag}") # used for get requests
    logger.debug(f"LSD token: {lsd_token}")
    logger.debug(f"jazoest: {jazoest}") # used for post requests
    logger.debug(f"jazoest async: {jazoest_async}") # used for get request
    logger.debug(f"client revision: {clientRevision}")
    logger.debug(f"client id(uuid): {clientID}")
    logger.debug(f"mqttAppID: {mqttAppID}")
    logger.debug(f"User App_ID: {userAppID}")        
    logger.debug(f"Mqtt Endpoint URL: {endpoint}")
    logger.debug(f"Region: {region}")
    return (fb_dtsg, fb_dtsg_ag, lsd_token, jazoest, jazoest_async, clientRevision, clientID, mqttAppID, userAppID, endpoint, region, user_name)



from typing import Any, Dict, List, Optional
from msgspec import Struct, json

from fbchat_muqit.models.attachment import StickerAttachment
from fbchat_muqit.models.thread import ThreadFolder, getThreadType
from .attachments_deltas import Mercury, extensibleattachment, BlobAttachment
from .parser import MessageParser
from ..message import Message, MessageReaction, MessageType, Reaction
from .custom_type import Value



class VNodes(Struct, frozen=True, eq=False):
    nodes: List

class VThreadMessages(Struct, frozen=True, eq=False):
    messages: VNodes 
    thread_key: Value
    thread_type: str
class VMessageThread(Struct, frozen=True, eq=False):
    message_thread: VThreadMessages

class VMessageAttachment(Struct, frozen=True, eq=False): 
    extensible_attachment: Optional[extensibleattachment] = None
    sticker: Optional[StickerAttachment] = None
    blob_attachments: Optional[List[BlobAttachment]] = None

def extract_thread_id(typ, obj)-> str:
    return Value(obj["thread_fbid"]) or Value(obj["other_user_id"])

thread_message_decoder = json.Decoder(type=List[VMessageThread], strict=False, dec_hook=extract_thread_id)


def parse_thread_message(payload: bytes)-> List[Message]:
    thread_message = thread_message_decoder.decode(payload)[0]
    thread_id = str(thread_message.message_thread.thread_key)
    nodes: List[Dict[str, Any]] = thread_message.message_thread.messages.nodes 

    messages = [
            Message(
                id=m["message_id"], 
                text=m["message"]["text"],
                sender_id=m["message_sender"]["id"],
                thread_id=thread_id,
                thread_type=getThreadType[thread_message.message_thread.thread_type],
                message_type=MessageType.TEXT,
                timestamp=m["timestamp_precise"],
                can_unsend=True if m["message_unsendability_status"] == "can_unsend" else False, 
                unsent=False,
                reaction=[
                    MessageReaction(
                        id="", 
                        reaction=r["reaction"],
                        reactor=r["user"]["id"],
                        thread_id=Value(thread_id),
                        reaction_type=Reaction.ADDED,
                        reacted_message_sender=m["message_sender"]["id"],
                        timestamp=None,
                    )   
                    for r in m["message_reactions"]
                ],
                mentions=None,
                thread_folder=ThreadFolder.INBOX,
                thread_participants=None,
                attachments=[MessageParser().parse_attachment(Mercury(
                    blob_attachment=m["blob_attachments"],
                    sticker_attachment=m["sticker"],
                    extensible_attachment=m["extensible_attachment"]
                    ))]
                )
            for m in nodes
            ]
    return messages


